#cimport csamtools

from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *
import pysam

import collections, array, struct, sys, itertools
import CGAT.Experiment as E
import CGAT.Intervals as Intervals
import CGAT.GTF as GTF
import CGAT.Stats as Stats
import numpy

CountResult = collections.namedtuple(
    "Counts", "upstream upstream_utr cds downstream_utr downstream")

class RangeCounter:
    
    def __init__(self, countfiles, 
                 controlfiles=None, 
                 control_factor=None, 
                 *args, **kwargs ):

        self.countfiles = countfiles
        self.controlfiles = controlfiles
        self.control_factor = control_factor 
        if self.control_factor is None:
            if self.controlfiles is not None:
                # count number of tags in each file for normalization purposes
                self.norm_fg, self.norm_bg = [], []
                for f in self.countfiles:
                    E.debug("computing read total for %s" % f)
                    self.norm_fg.append(self.getTotal(f))
                for f in self.controlfiles:
                    E.debug("computing read total for %s" % f)
                    self.norm_bg.append(self.getTotal(f))
                fg = sum(self.norm_fg)
                bg = sum(self.norm_bg)
                if bg != 0:
                    self.control_factor = float(fg) / float(bg)
                else:
                    raise ValueError('no background data for normalization')
                E.info("normalization: fg=%i, bg=%f, factor=%f" %
                       (fg, bg, self.control_factor))

    def setup(self, ranges):
        self.counts = numpy.zeros(Intervals.getLength(ranges))
        if self.controlfiles:
            self.counts_bg = numpy.zeros(Intervals.getLength(ranges))

    def getTotal(self):
        raise NotImplementedError(
            'implementations of RangeCounter need to implement getTotal')

    def getCounts(self, contig, ranges, length=0):
        '''count from a set of ranges.

        The ranges are scaled towards the length of the intervals that
        are being counted.
        '''

        self.setup(ranges)
        self.count(self.counts, self.countfiles, contig, ranges)
        if self.controlfiles:
            self.count(self.counts_bg, self.controlfiles, contig, ranges)

        # subsample for length
        if length > 0:
            lnormed = float(length)
            lunnormed = float(len(self.counts))
            if lunnormed == 0: return numpy.zeros(0)
            # E.debug( "normed length = %i, unnormed length =%i " % (lnormed, lunnormed) )
            if lunnormed > lnormed:
                # compress by taking only counts at certain points within intervals
                take = numpy.unique( numpy.array( numpy.floor( 
                            numpy.arange( 0, lunnormed - 1, lunnormed / lnormed ) ), dtype = int ) )
            elif lunnormed == lnormed:
                # same size, no scaling
                take = numpy.arange( 0, lnormed, dtype = int )
            else:
                # expand by taking more counts
                take = numpy.array( numpy.arange(0, lunnormed, lunnormed / lnormed ), dtype = int )[:length]
                
            assert len(take) == length, "size of arrays unequal: %i != %i from %i" % (len(take), length, lunnormed)

            self.counts = self.counts[take]
            if self.controlfiles:
                self.counts_bg = self.counts_bg[take]

        if self.controlfiles:
            # scale background 
            self.counts += 1
            self.counts_bg += 1
            self.counts = numpy.array( self.counts, dtype=float) * \
                self.control_factor / self.counts_bg

        return self.counts

class RangeCounterBAM(RangeCounter):
    '''count densities using bam files.
    '''

    def getTotal(self, bamfile):
        '''return total number of tags in bedfile.'''
        return bamfile.mapped

    def __init__(self, *args, **kwargs ):
        '''
        :param samfiles: list of :term:`bam` formatted files
        :param shifts: list of read shifts
        :param extends: list of read extensions
        :param merge_pairs: merge read pairs for counting
        :param min_insert_size: remove paired reads with insert size below this threshold
        :param max_insert_size: remove paired reads with insert size above this threshold
        '''

        RangeCounter.__init__(self, *args, **kwargs )

    def count(self, counts, files, contig, ranges):

        if len(ranges) == 0: return

        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int xstart, xend, rstart, rend, start, end
        cdef int interval_width
        cdef int current_offset
        cdef AlignedSegment read
        cdef int shift_extend
        cdef int work_offset
        cdef int pos
        cdef int length

        cdef AlignmentFile samfile

        for samfile in files:

            current_offset = 0

            for start, end in ranges:
                length = end - start

                # skip bamfiles where contig is not present
                if samfile.gettid(contig) < 0:
                    continue

                for read in samfile.fetch(contig, start, end):
                    rstart = max(start, read.pos) - start + current_offset
                    # skip unmapped reads that are assigend a position.
                    if read.aend is None:
                        continue
                    rend = min( end, read.aend) - start + current_offset
                    for i from rstart <= i < rend: counts[i] += 1

                current_offset += length

    def getTotal(self, samfile):
        '''return total number of mapped tags in samfile.'''
        return samfile.mapped


class RangeCounterBAMShift(RangeCounterBAM):
    '''count densities using bam files. 

    Before counting, reads are shifted and extended by a fixed amount.
    '''
    def __init__(self, countfiles, shifts, extends, 
                 controlfiles = None,
                 *args, **kwargs ):
        '''
        :param samfiles: list of :term:`bam` formatted files
        :param shifts: list of read shifts
        :param extends: list of read extensions
        '''

        RangeCounterBAM.__init__(self, countfiles,
                                 controlfiles = controlfiles, *args, **kwargs)

        self.shifts, self.extends = [], []

        for x in xrange(max( (len(shifts), len(extends), len(countfiles)) )):
            if x >= len(shifts):
                shift = 0
            else:
                shift = shifts[0]
            if x >= len(extends):
                extend = 0
            else:
                extend = extends[0]

            if shift > 0 and extend == 0:
                E.warn( "no extension given for shift > 0 - extension will be 2 * shift" )
                extend = shift * 2

            self.shifts.append( shift )
            self.extends.append( extend )

    def count(self, counts, files, contig, ranges ):

        if len(ranges) == 0: return

        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int xstart, xend, rstart, rend, start, end
        cdef int interval_width
        cdef int current_offset
        cdef AlignedSegment read
        cdef int shift_extend
        cdef int work_offset
        cdef int pos
        cdef int length

        # shifting:
        # forward strand reads:
        #   - shift read upstream by shift
        #   - extend read upstream by extend 
        # reverse strand: 
        #   - shift read downstream by shift
        #   - extend read downstream by extend

        # There is a problem of discontinuity through shifting at
        # exon boundaries
        # 1. Density will accumulate at the ends as reads are shifted
        # up and downstream. However, shifted density will end up
        # in the introns and thus not counted. 
        # 2. The densities along exon boundaries will always be 
        # discontinuous.

        cdef int extend
        cdef int shift
        cdef AlignmentFile samfile

        for samfile, shift, extend in zip(files, self.shifts, self.extends):

            current_offset = 0
            shift_extend = shift + extend

            for start, end in ranges:
                length = end - start
                # collect reads including the regions left/right of interval
                xstart, xend = max(0, start - shift_extend), max(0, end + shift_extend)

                # skip bamfiles where contig is not present
                if samfile.gettid(contig) < 0:
                    continue

                for read in samfile.fetch(contig, xstart, xend):
                    if read.is_reverse: 
                        rstart = read.aend - start - shift_extend
                    else:
                        rstart = read.pos - start + shift

                    rend = min( length, rstart + extend ) + current_offset
                    rstart = max( 0, rstart ) + current_offset
                    for i from rstart <= i < rend: counts[i] += 1

                current_offset += length

class RangeCounterBAMMerge(RangeCounterBAM):
    '''count densities using bam files.

    Before counting, paired end reads are merged and counts
    are computed over the full merged range.

    Reads are not shifted.
    '''
    def __init__(self, countfiles,
                 merge_pairs, 
                 min_insert_size, max_insert_size, 
                 controlfiles = None,
                 *args, **kwargs ):
        '''
        :param samfiles: list of :term:`bam` formatted files
        :param min_insert_size: remove paired reads with insert size below this threshold
        :param max_insert_size: remove paired reads with insert size above this threshold
        '''

        RangeCounterBAM.__init__(self, countfiles,
                                 controlfiles = controlfiles,
                                 *args, **kwargs )
        self.merge_pairs = merge_pairs
        self.min_insert_size = min_insert_size
        self.max_insert_size = max_insert_size

    def count(self, counts, files, contig, ranges ):

        if len(ranges) == 0:
            return

        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int xstart, xend, rstart, rend, start, end
        cdef int interval_width
        cdef int current_offset
        cdef AlignedSegment read
        cdef int work_offset
        cdef int pos
        cdef int length

        cdef AlignmentFile samfile
        cdef int min_insert_size = self.min_insert_size
        cdef int max_insert_size = self.max_insert_size

        for samfile in files:

            current_offset = 0

            for start, end in ranges:
                length = end - start
                
                xstart, xend = start, end

                # skip bamfiles where contig is not present
                if samfile.gettid(contig) < 0:
                    continue

                for read in samfile.fetch(contig, xstart, xend):
                    flag = read._delegate.core.flag 
                    # remove unmapped reads
                    if flag & 4:
                        continue
                    # remove unpaired
                    if not flag & 2:
                        continue
                    # this is second pair of read - skip to avoid double counting
                    if flag & 128:
                        continue
                    # remove reads on different contigs
                    if read.tid != read.mrnm:
                        continue
                    # remove if insert size too large
                    if (read.isize > max_insert_size) or \
                       (read.isize < min_insert_size):
                        continue
                    if read.pos < read.mpos:
                        rstart = max(start, read.pos)
                        rend = min(end, read.mpos + read.rlen)
                    else:
                        rstart = max(start, read.mpos)
                        rend = min(end, read.pos + read.rlen)

                    rstart += -start + current_offset
                    rend += -start + current_offset
                    for i from rstart <= i < rend:
                        counts[i] += 1

                current_offset += length

class RangeCounterBAMBaseAccuracy(RangeCounterBAM):
    '''count densities using bam files with base accuracy.
    '''
    def __init__(self, 
                 *args, **kwargs):
        '''
        '''

        RangeCounterBAM.__init__(self, *args, **kwargs )

    def count(self, counts, files, contig, ranges):

        if len(ranges) == 0:
            return

        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int xstart, xend, rstart, rend, start, end
        cdef int interval_width
        cdef int current_offset
        cdef AlignedSegment read
        cdef int work_offset
        cdef int pos
        cdef int length
        
        cdef AlignmentFile samfile

        for samfile in files:

            current_offset = 0

            for start, end in ranges:
                length = end - start
                # skip bamfiles where contig is not present
                if samfile.gettid(contig) < 0:
                    continue

                for read in samfile.fetch(contig, start, end):
                    # note: not optimized for speed yet - uses a python list
                    positions = read.positions
                    for i in positions:
                        if i < start: continue
                        if i >= end: continue
                        try:
                            counts[i - start + current_offset] += 1
                        except IndexError:
                            print len(counts)
                            print "i=", i, "start=",start, "end=", end, "current_offset=", current_offset
                            print "positions=", positions
                            raise

                current_offset += length

class RangeCounterBed(RangeCounter):

    def __init__(self, *args, **kwargs):
        RangeCounter.__init__(self, *args, **kwargs)
                
    def getTotal(self, bedfile):
        '''return total number of tags in bedfile.'''
        cdef int total = 0
        for x in bedfile.fetch():
            total += 1
        return total

    def count(self, counts, files, contig, ranges):
        
        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int xstart, xend, rstart, rend, start, end

        if len(ranges) == 0: return

        cdef int length
        cdef int current_offset
        cdef int interval_width

        for bedfile in files:
            current_offset = 0

            for start, end in ranges:
                length = end - start
                try:
                    for bed in bedfile.fetch(contig, max(0, start), end,
                                             parser = pysam.asBed()):
                        # truncate to range of interest
                        rstart = max(0, bed.start - start) + current_offset
                        rend = min( length, bed.end - start) + current_offset
                        for i from rstart <= i < rend: counts[i] += 1
                except (ValueError, KeyError):
                    # contig not present
                    pass

                current_offset += length

        

class RangeCounterBigWig(RangeCounter):

    def __init__(self, *args, **kwargs ):
        RangeCounter.__init__(self, *args, **kwargs )
        
    def getTotal(self, filename):
        raise NotImplementedError(
            "getTotal not implemented for RangeCounterBigwig")

    def count(self, counts, files, contig, ranges ):
        
        # collect pileup profile in region bounded by start and end.
        cdef int i
        cdef int rstart, rend, start, end, tstart, tend

        if len(ranges) == 0:
            return

        cdef int length
        cdef int current_offset

        for wigfile in files:

            current_offset = 0
            for start, end in ranges:
                length = end - start

                values = wigfile.get(contig, start, end)
                if values == None:
                    continue

                for tstart, tend, value in values:
                    rstart = tstart - start + current_offset
                    rend = tend - start + current_offset
                    for i from rstart <= i < rend: counts[i] = value
                    
                current_offset += length

class IntervalsCounter:
    '''aggregate reads/intervals within a certain arrangement of
    intervals (sections). Sections are defined by a transcript-model 
    and might be configurations such as upstream/exon/downstream, 
    exon/intron, etc.

    Derived class should provide a ``self.count()`` method that accepts
    a list of intervals within a transcript and gene model,
    collects counts within some configuration based on the
    transcript structure and then calls ``self.aggregate()`` to
    aggregate these counts.
    
    The base class provides normalization and book keeping methods
    for the aggregation process.
    '''

    format = "%i"

    def __init__(self, counter, 
                 normalization = None, 
                 outfile_profiles = None,
                 *args, **kwargs ):
        '''counter is an object to provide counts in a list of regions
        scaled to a common size.
        '''

        self.counter = counter

        # normalization method to use before aggregation
        self.normalization = normalization        
        # lables of categories to count (introns, exons, ...)
        self.fields = []
        # number of items counted in each category (1012 introns, 114 exons, ...)
        self.counts = []
        # lengths of items counted in each category
        self.lengths = []
        # number of bins in this category
        self.nbins = []
        # aggregate counts
        self.aggregate_counts = []
        # number of aggregates skipped due to shape mismatch
        # (happens if region extends beyound start/end of contig
        self.nskipped = 0

        # for keeping totals without aggregation
        self.outfile_profiles = outfile_profiles

    def add(self, field, length):
        '''add a new region to counter.

        The region is named *field* and has length *length*.

        '''
        self.counts.append(0)
        self.lengths.append([])
        self.fields.append(field)
        self.nbins.append(length)
        if self.normalization in ("max", "sum", "total-max", "total-sum"):
            self.dtype = numpy.float
        else:
            self.dtype = numpy.int
        self.aggregate_counts.append(numpy.zeros(length, dtype=self.dtype))

    def setOutputProfiles(self, outfile_profiles):
        if outfile_profiles:
            self.outfile_profiles = outfile_profiles
            sum_counts = sum([len(x) for x in self.aggregate_counts])
            self.outfile_profiles.write(
                "\t%s\n" % ("\t".join(map(str, range(sum_counts)))))

    def getNumBins( self ):
        '''returns the number of bins in this counter.'''
        return sum(self.nbins )

    def closeOutputProfiles( self ):
        if self.outfile_profiles:
            self.outfile_profiles.close()

    def setNormalization( self, normalization ):
        self.normalization = normalization

        if self.normalization in ("max", "sum", "total-max", "total-sum" ):
            self.dtype = numpy.float
            self.format = "%6.4f"
        else:
            self.dtype = numpy.int
            self.format = "%i"

        for x,c in enumerate(self.aggregate_counts):
            self.aggregate_counts[x] = numpy.zeros( len(c), dtype = self.dtype ) 

    def addName( self, name ):
        '''record name currently being processed.'''
        self.transcript_names.append( name )

    def addLengths( self, *lengths ):
        '''add interval lengths to this counter.'''

        assert len(lengths) == len(self.lengths)
        for idx, interv in enumerate(lengths):
            self.lengths[idx].append(sum([x[1]-x[0] for x in interv]))

    def aggregate(self, *counts):

        # save for outputting
        self.last_counts = counts

        # normalization across all regions
        if self.normalization == "total-max":
            norm = float(max([max(c) for c in counts if len(c) > 0]))
        elif self.normalization == "total-sum":
            norm = float(sum([sum(c) for c in counts] ))
        else:
            norm = 0

        for x, xx in enumerate( zip( self.aggregate_counts, counts)):
            agg, c = xx
            self.counts[x] += 1

            if len(c) == 0:
                continue
            
            if norm > 0:
                # apply uniform normalization
                cc = c / norm
            elif self.normalization == "max":
                # normalize by max count per region
                m = max(c)
                if m == 0: continue
                cc = c / m
            elif self.normalization == "sum":
                # normalize by sum counts per region
                m = sum(c)
                if m == 0: continue
                cc = c / m
            else:
                # no normalization
                cc = c

            try:
                agg += cc.astype(agg.dtype)
            except ValueError:
                self.nskipped += 1

    def getProfile(self,
                   normalize=None,
                   background_region_bins=10):
        '''return meta gene profile.

        apply various normalization methods.

        area
           Normalize by total area under curve of
           the complete profile.

        counts
           Normalize each segment individually by
           the number of regions included.
        
        background
           Apply background normalization to complete
           profile.

        '''
        if normalize is None or normalize == "none":
            profile = numpy.concatenate(self.aggregate_counts)
        
        elif normalize in ("area", "background"):
            profile = numpy.concatenate(self.aggregate_counts)
            profile = numpy.array(profile, dtype=numpy.float)
            
            if normalize == "area":
                background = profile.sum()
                
            elif normalize == "background":
                # take counts at either end of profile
                background_counts = numpy.concatenate( 
                    [profile[:background_region_bins],
                     profile[-background_region_bins:]])
                # compute mean()
                background = numpy.mean(background_counts)

            profile /= background

        elif normalize == "counts":
            profile = numpy.concatenate(
                [numpy.array(x, dtype=numpy.float) / y
                 for x, y in zip(self.aggregate_counts,
                                 self.counts)])

        return profile

    def writeLengthStats(self, outfile):
        '''output length stats to outfile.'''
        
        outfile.write("region\t%s\n" % "\t".join(Stats.Summary.fields))
        for field, l in zip(self.fields, self.lengths):
            outfile.write("%s\t%s\n" % (field, str(Stats.Summary(l))))

    def update(self, gtf):
        
        counted = self.count(gtf)

        if counted and self.outfile_profiles:
            name = gtf[0].transcript_id
            self.outfile_profiles.write("%s\t%s\n" % (name,
                                        "\t".join( [ "\t".join( map(str, x) ) for x in self.last_counts ] ) ) )
        
        return counted

    def __str__(self):
        return "%s=%s" % (self.name, ",".join( [str(sum(x)) for x in self.aggregate_counts]) )

class UnsegmentedCounter(IntervalsCounter):
    '''clone a segmented counter as a single
    unsegmented counter.
    '''
    
    name = None
    
    def __init__(self, counter,
                 *args,
                 **kwargs ):
        IntervalsCounter.__init__(self, counter, *args, **kwargs)
        self.name = counter.name
        self.add( 'all', counter.getNumBins() )


class SeparateExonCounter(IntervalsCounter):
    '''count reads over 1st, middle and last exons as well as upstream/
    downstream of genes/transcripts.  Only protein-coding genes (with a
    CDS) are counted
    '''
    name = "separateexonprofile"

    def __init__(self, counter, 
                 int resolution_upstream,
                 int resolution_first_exon,
                 int resolution_last_exon,
                 int resolution_exons,
                 int resolution_downstream,
                 int extension_upstream = 0,
                 int extension_downstream = 0,
                 int scale_flanks = 0,
                 *args,
                 **kwargs):

        IntervalsCounter.__init__(self, counter, *args, **kwargs)
	      
        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream
        self.resolution_first_exon = resolution_first_exon
        self.resolution_last_exon = resolution_last_exon
        self.resolution_exons = resolution_exons
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        self.scale_flanks = scale_flanks

        for field, length in zip(
                ("upstream", "first_exon", "exons", "last_exon", "downstream"),
                (resolution_upstream,
                 resolution_first_exon,
                 resolution_exons,
                 resolution_last_exon,
                 resolution_downstream)):
            self.add(field, length)


    def count(self, gtf):
        '''build ranges to be analyzed from a gene model.
        Returns a tuple with ranges for first, middle and last exons,
        upstream and downstream.  Does not count genes with 1 exon.
        '''
        
        contig = gtf[0].contig
        exons = GTF.asRanges(gtf, "exon")
        if len(exons) == 0:
            E.warn("no exons in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        if len(exons) == 2:
            first = [exons[0]]
            last = [exons[-1]]
            exons = [(0, 0)]

        elif len(exons) == 1:
            E.warn("only one exon in %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        else:
            first = [exons[0]]
            last = [exons[-1]]
            exons.remove(exons[0])
            exons.remove(exons[-1])

        first_start, first_end = first[0][0], first[0][1]
        last_start, last_end = last[0][0], last[0][1]
        exons_start, exons_end = exons[0][0], exons[-1][1]
        if self.scale_flanks > 0:
            self.extension_downstream = (last_end - first_start)*self.scale_flanks
            self.extension_upstream = (last_end - first_start)*self.scale_flanks
            E.dubg("scale flanks")
            
        if gtf[0].strand == "-":
            downstream = [(max(0, first_start - self.extension_downstream),
                           first_start),]
            upstream = [(last_end,
                         last_end + self.extension_upstream), ]
        else:
            upstream = [(max(0, first_start - self.extension_upstream),
	                 first_start), ]
            downstream = [(last_end,
	                   last_end + self.extension_downstream), ]

        E.debug("counting first exons")
        self.counts_first_exon = self.counter.getCounts(contig, first,
                                                        self.resolution_first_exon)

        E.debug("counting exons")
        self.counts_exons = self.counter.getCounts(contig,exons,
                                                   self.resolution_exons)

        E.debug("counting last exons")
        self.counts_last_exon = self.counter.getCounts(contig, last,
                                                       self.resolution_last_exon)

        E.debug("counting upstream")
        self.counts_upstream = self.counter.getCounts(contig, upstream,
                                                      self.resolution_upstream)

        E.debug("counting downstream")
        self.counts_downstream = self.counter.getCounts(contig, downstream,
                                                        self.resolution_downstream)
        
        E.debug("counting finished" )
        
        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_first_exon = self.counts_first_exon[::-1]
            self.counts_last_exon = self.counts_last_exon[::-1]
            self.counts_exons = self.counts_exons[::-1]
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]

        self.addLengths( upstream, first, exons, last, downstream )

        self.aggregate(self.counts_upstream,
                       self.counts_first_exon,
                       self.counts_exons,
                       self.counts_last_exon,
                       self.counts_downstream )

        return 1


class SeparateExonWithIntronCounter(IntervalsCounter):
    '''count reads over 1st, middle and last exons as well as introns, upstream/
    downstream of genes/transcripts.  Only protein-coding genes (with a
    CDS) are counted, introns are not split.
    '''
    name = "separateexonprofilewithintrons"

    def __init__(self, counter, 
                 int resolution_upstream,
                 int resolution_first_exon,
                 int resolution_last_exon,
                 int resolution_exons,
                 int resolution_introns,
                 int resolution_downstream,
                 int extension_upstream = 0,
                 int extension_downstream = 0,
                 int scale_flanks = 0,
                 *args,
                 **kwargs):

        IntervalsCounter.__init__(self, counter, *args, **kwargs)
	      
        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream
        self.resolution_first_exon = resolution_first_exon
        self.resolution_last_exon = resolution_last_exon
        self.resolution_exons = resolution_exons
        self.resolution_introns = resolution_introns
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        self.scale_flanks = scale_flanks

        for field, length in zip(
                ("upstream", "first_exon", "exons",
                 "introns", "last_exon", "downstream"),
                (resolution_upstream,
                 resolution_first_exon,
                 resolution_exons,
                 resolution_introns,
                 resolution_last_exon,
                 resolution_downstream)):
            self.add(field, length)

    def count(self, gtf):
        '''build ranges to be analyzed from a gene model.
        Returns a tuple with ranges for first, middle and last exons, introns,
        upstream and downstream.  Does not count genes with 1 exon.
        '''
        
        contig = gtf[0].contig
        exons = GTF.asRanges(gtf, "exon")
        introns = Intervals.complement(exons)

        if len(exons) == 0:
            E.warn("no exons in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        # Exclude intron-less genes
        if len(introns) == 0:
            E.warn("no introns in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0
	  
        if len(exons) == 2:
            first = [exons[0]]
            last = [exons[-1]]
            exons = [(0, 0)]

        elif len(exons) == 1:
            E.warn("only one exon in %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        else:
            first = [exons[0]]
            last = [exons[-1]]
            exons.remove(exons[0])
            exons.remove(exons[-1])

        first_start, first_end = first[0][0], first[0][1]
        last_start, last_end = last[0][0], last[0][1]
        exons_start, exons_end = exons[0][0], exons[-1][1]
        if self.scale_flanks > 0:
            self.extension_downstream = (last_end - first_start)*self.scale_flanks
            self.extension_upstream = (last_end - first_start)*self.scale_flanks
            E.dubg("scale flanks")
            
        if gtf[0].strand == "-":
            downstream = [(max(0, first_start - self.extension_downstream),
                           first_start),]
            upstream = [(last_end,
                         last_end + self.extension_upstream), ]
        else:
            upstream = [(max(0, first_start - self.extension_upstream),
	                 first_start), ]
            downstream = [(last_end,
	                   last_end + self.extension_downstream), ]

        E.debug("counting first exons")
        self.counts_first_exon = self.counter.getCounts(contig, first,
                                                        self.resolution_first_exon)

        E.debug("counting exons")
        self.counts_exons = self.counter.getCounts(contig, exons,
                                                   self.resolution_exons)

        E.debug("counting introns")
        self.counts_introns = self.counter.getCounts(contig, introns,
                                                     self.resolution_introns)

        E.debug("counting last exons")
        self.counts_last_exon = self.counter.getCounts(contig, last,
                                                       self.resolution_last_exon)

        E.debug("counting upstream")
        self.counts_upstream = self.counter.getCounts(contig, upstream,
                                                      self.resolution_upstream)

        E.debug("counting downstream")
        self.counts_downstream = self.counter.getCounts(contig, downstream,
                                                        self.resolution_downstream)
        
        E.debug("counting finished" )
        
        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_first_exon = self.counts_first_exon[::-1]
            self.counts_last_exon = self.counts_last_exon[::-1]
            self.counts_exons = self.counts_exons[::-1]
            self.counts_introns = self.counts_introns[::-1]
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]

        self.addLengths(upstream, first, exons, introns, last, downstream)

        self.aggregate(self.counts_upstream,
                       self.counts_first_exon,
                       self.counts_exons,
                       self.counts_introns,
                       self.counts_last_exon,
                       self.counts_downstream )

        return 1


class GeneCounter( IntervalsCounter ):
    '''count reads in exons and upstream/downstream of genes/transcripts.
    
    Only protein coding transcripts are counted (i.e. those with a CDS)
    '''
    
    name = "geneprofile"
    
    def __init__(self, counter, 
                 int resolution_upstream,
                 int resolution_exons,
                 int resolution_downstream,
                 int extension_upstream = 0, 
                 int extension_downstream = 0,
                 int scale_flanks = 0,
                 *args,
                 **kwargs ):

        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream 
        self.resolution_exons = resolution_exons
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        self.scale_flanks = scale_flanks

        for field, length in zip( 
            ("upstream", "exons", "downstream"),
            (resolution_upstream,
             resolution_exons,
             resolution_downstream ) ):
            self.add( field, length )
        
    def count( self, gtf ):
        '''build ranges to be analyzed from a gene model.

        Returns a tuple with ranges for exons, upstream, downstream.
        '''

        contig = gtf[0].contig 
        exons = GTF.asRanges( gtf, "exon" )
        if len(exons) == 0:
            E.warn( "no exons in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        exon_start, exon_end = exons[0][0], exons[-1][1]
        if self.scale_flanks > 0:
            self.extension_downstream = (exon_end - exon_start)*self.scale_flanks
            self.extension_upstream = (exon_end - exon_start)*self.scale_flanks
            E.debug("scale flanks")

        if gtf[0].strand == "-":
            downstream = [( max(0, exon_start - self.extension_downstream),
                            exon_start ),] 
            upstream = [( exon_end,
                          exon_end + self.extension_upstream),]
        else:
            upstream = [(max(0, exon_start - self.extension_upstream),
                         exon_start ),] 
            downstream = [(exon_end,
                           exon_end + self.extension_downstream ),]

        E.debug("counting exons" )
        self.counts_exons = self.counter.getCounts(contig,
                                                   exons,
                                                   self.resolution_exons)
        E.debug("counting upstream" )
        self.counts_upstream = self.counter.getCounts(contig, 
                                                      upstream,
                                                      self.resolution_upstream) 
        E.debug("counting downstream" )
        self.counts_downstream = self.counter.getCounts(contig,
                                                        downstream,
                                                        self.resolution_downstream)

        E.debug("counting finished" )

        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_exons = self.counts_exons[::-1]
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]
            
        self.addLengths( upstream, exons, downstream )

        self.aggregate( self.counts_upstream,
                        self.counts_exons,
                        self.counts_downstream )

        return 1


class GeneCounterWithIntrons( IntervalsCounter ):
    '''count reads in exons, in introns and the upstream/downstream of genes/transcripts.
    
    Assumes all the behavior of GeneCounter class, because the code is modified based on 
    GeneCounter Class. Class created by Tim on 16 May. 2013.
    
    Both  protein coding transcripts, and non coding transcripts are counted.
    (i.e. both those with a CDS, and those without a CDS are counted)
    '''
    
    name = "geneprofilewithintrons"
    
    def __init__(self, 
                 counter,
                 int resolution_upstream,
                 int resolution_exons,
                 int resolution_introns,  #Tim 16 May. 2013. for introns
                 int resolution_downstream,
                 int extension_upstream = 0, 
                 int extension_downstream = 0,
                 int scale_flanks = 0,
                 *args,
                 **kwargs ):

        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream 
        self.resolution_exons = resolution_exons
        self.resolution_introns = resolution_introns   #Tim 16 May. 2013. for introns
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        self.scale_flanks = scale_flanks

        for field, length in zip( 
            ("upstream", "exons", "introns", "downstream"),  #Tim 16 May. 2013. for introns
            (resolution_upstream,
             resolution_exons,
             resolution_introns,  #Tim 16 May. 2013. for introns
             resolution_downstream ) ):
            self.add( field, length )
        
    def count( self, gtf ):
        '''build ranges to be analyzed from a gene model.

        Returns a tuple with ranges for upstream, exons, introns, downstream.
        '''

        contig = gtf[0].contig 
        exons = GTF.asRanges( gtf, "exon" )
        introns = Intervals.complement(exons) #Tim 16 May. 2013. for introns
        
        if len(exons) == 0:
            E.warn( "no exons in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0
            
        # Tim 16 May. 2013. for introns. 
        # Genes have no introns are omitted for counting in all other regions. Because 
        # include them will distort the relative height of exons/introns in the profile.
        # In most gene profile plots, the relative height of different regions 
        # is very important info, thus omit the gene will ensure the correctness for the 
        # relative height across different regions.
        # Particularly true when user use normalization method: total-max.
        if len(introns) == 0:   
            E.debug( "Omitted this gene entirely because there are no introns in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0

        exon_start, exon_end = exons[0][0], exons[-1][1]
        if self.scale_flanks > 0:
            self.extension_downstream = (exon_end - exon_start)*self.scale_flanks
            self.extension_upstream = (exon_end - exon_start)*self.scale_flanks
            E.debug("scale flanks")

        if gtf[0].strand == "-":
            downstream = [ ( max(0, exon_start - self.extension_downstream), exon_start ), ] 
            upstream = [ ( exon_end, exon_end + self.extension_upstream ), ]
        else:
            upstream = [ ( max(0, exon_start - self.extension_upstream), exon_start ), ] 
            downstream = [ ( exon_end, exon_end + self.extension_downstream ), ]

        E.debug("counting exons" )
        self.counts_exons = self.counter.getCounts( contig, exons, self.resolution_exons  )
        E.debug("counting introns" ) #Tim 16 May. 2013. for introns. 
        self.counts_introns = self.counter.getCounts( contig, introns, self.resolution_introns  )        
        E.debug("counting upstream" )
        self.counts_upstream = self.counter.getCounts( contig, upstream, self.resolution_upstream ) 
        E.debug("counting downstream" )
        self.counts_downstream = self.counter.getCounts( contig, downstream, self.resolution_downstream )

        E.debug("counting finished" )

        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_exons = self.counts_exons[::-1]
            self.counts_introns = self.counts_introns[::-1]   #Tim 16 May. 2013. for introns. 
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]
            
        self.addLengths( upstream, exons, introns, downstream )   #Tim 16 May. 2013. for introns. 

        self.aggregate( self.counts_upstream,
                        self.counts_exons,
                        self.counts_introns,  #Tim 16 May. 2013. for introns. 
                        self.counts_downstream )

        return 1



class GeneCounterAbsoluteDistanceFromThreePrimeEnd( IntervalsCounter ):
    '''count reads in exons of genes/transcripts from the three prime
    polyA tail, without shrinking or lengthening the gene as it would
    in other genecounter mode in this script.  

    In other words, only count the reads that fall on the exons of the
    (virtual maximal) mRNA transcript.  Note that the distance is
    relative to TTS (three prime polyA tail) on the mNRA transcript,
    insteads of on the genomic assembly is used for the counting and
    plotting.
    
    For mRNA with multiple exons, the exons are first stiched together
    into one piece of mRNA (the virtual maximal transcript for each
    gene). Subsequently, the mRNA is being used for counting. This is
    the (only) proper way to avoid counting in introns, which screw up
    the actual genebody coverage profile. And this only works with
    --base-accuracy option, because otherwise spliced reads will be
    considered as covering the entire intron, or spliced out introns
    will be counted as covered by reads.  (this option imply the
    --base-accuracy option).
    
    Also count reads in introns (in exactly the same manner as if they
    are exons described above ) of genes/transcripts from the three
    prime polyA tail.
    
    Note:

    * Both protein coding transcripts and non coding
    transcripts are counted.  i.e. both those with a CDS, and those
    without a CDS are counted.
    
    * Only genes/transcripts with total length longer than the
    --extension-exons-absolute-distance-topolya is counted (i.e. those
    genes/transcripts shorter than this is omitted because they will
    only contribute to the last len(gene) base pair of the graph, thus
    end-up ploting a graph appears to have three prime bias even when
    there is no three prime bias at all.
    
    There is one alternative way of dealing with this issue: is to
    count the number of times a nucleotide is being visited, and
    normalize against the counts curve with the visits counting
    curve. However, under current architecture of the counter, it
    seems hard to support this method.  Assumeing you get enough
    number of genes longer than 4000b (in hg19) to plot the three
    prime bias to satisfactory quality, it seems not really worth the
    effort to implement the alternative way,

    '''
    
    name = "geneprofileabsolutedistancefromthreeprimeend"   #Tim 31th Aug 2013
    
    def __init__(self, 
                 counter,
                 int resolution_upstream,
                 int resolution_downstream,
                 int resolution_exons_absolute_distance_topolya,
                 int resolution_introns_absolute_distance_topolya,
                 # int resolution_exons_absolute_distance_tostartsite,   # Tim 31th Aug 2013: 
                                                                         # a possible feature for future,  if five prime bias is of your interest. 
                                                                         #(you need to create another class). It is not very difficult to derive from this class, but is not implemented yet
                                                                         # This future feature is slightly different the TSS profile already implemented, because in this future feature introns are skipped, 
                 # int resolution_introns_absolute_distance_tostartsite, # feature not implemented yet
                 
                 int extension_upstream = 0, 
                 int extension_downstream = 0,
                 int extension_exons_absolute_distance_topolya = 0,
                 int extension_introns_absolute_distance_topolya = 0,
                 # int extension_exons_absolute_distance_tostartsite = 0,   # feature not implemented yet
                 # int extension_introns_absolute_distance_tostartsite = 0, # feature not implemented yet
                 
                 int scale_flanks = 0,
                 *args,
                 **kwargs ):

        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream 
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        
        self.resolution_exons_absolute_distance_topolya = resolution_exons_absolute_distance_topolya     #Tim 31th Aug 2013
        self.resolution_introns_absolute_distance_topolya = resolution_introns_absolute_distance_topolya #Tim 31th Aug 2013
        self.extension_exons_absolute_distance_topolya = extension_exons_absolute_distance_topolya       #Tim 31th Aug 2013
        self.extension_introns_absolute_distance_topolya = extension_introns_absolute_distance_topolya   #Tim 31th Aug 2013
        
        self.scale_flanks = scale_flanks

        for field, length in zip(
                ("upstream%dbp_zoomedTo%dbp"%(extension_upstream, resolution_upstream),
                 "exonsLast%dbp_zoomedTo%dbp"%(extension_exons_absolute_distance_topolya, resolution_exons_absolute_distance_topolya) ,
                 "intronsLast%dbp_zoomedTo%dbp"%(extension_introns_absolute_distance_topolya, resolution_introns_absolute_distance_topolya) ,
                 "downstream%dbp_zoomedTo%dbp"%(extension_downstream, resolution_downstream),
             ),
                (resolution_upstream,
                 resolution_exons_absolute_distance_topolya,
                 resolution_introns_absolute_distance_topolya,
                 resolution_downstream) 
        ):
            self.add(field, length)

            
    #Tim 31th Aug 2013: Important function in this class to stich together the exons into one complete of mRNA
    def __chopAllTranscriptToAFixedLengthForTheThreePrimeBiasCounting(self, gtf , exons, extension_threePrimeBiasNotZoomed ):
        '''
        works for both exons and introns, the variable name exons is only symolic here,
        it means either exons, or introns depending on the regions you passed in 
        during function call
        
        in order to use the gene2profile code to plot 3prime bias plot, 
        if a gene is longer than my plotting size (e.g. 5kb), because I 
        do not want to let the code to shrink the gene, I need to create 
        a "chopped gene" that only has the first several bp (e.g. 5kb) 
        closest to the 3prime, so the gene always appears to be of the length 
        "extension_threePrimeBiasNotZoomed", (e.g. 5kb)
        
        In this version, shorter genes are omitted at count() becase len([])==0.
        It is oimitted in the same mannar that was used to
        omit the genes without any exons, by returnning [], which can be interpreted
        by the calling function as omit, or take any other further actions.
       
        '''
        
        if len(exons)==0: return []  # this is to cater for the case where there's 
                                     # one exon, and no intron, but I am plotted\
                                     # unzoomed introns as well
        
        leftoverBasepairAllowance = extension_threePrimeBiasNotZoomed
        exons_chopped = []
        if gtf[0].strand == "-":
            for exon in exons:
                exonSize = exon[1]-exon[0]
                if leftoverBasepairAllowance > exonSize:
                    leftoverBasepairAllowance -= exonSize
                    exons_chopped.append(exon)            # append is the correct way, coz we process the first exon first. 
                                                            # so subsequent exons shall be appended so that the basepair ordering 
                                                            # is in reverse ordering alone the gene body, likewise, the exon 
                                                            # ordering is also the reverse ordering alone the gene body. In the 
                                                            # end, just before counts are aggregated, there will be code:
                                                            # if "-": counts_exons_NotZoomedAndChopped_ForThreePrimeBias[::-1] to flip 
                                                            # everything in base pair resolution.
                    E.debug(str(exons_chopped))
                else:
                    E.debug("Exon intervals preparation function: gene %s:%s, leftoverBasepairAllowance= %s " % (gtf[0].transcript_id, gtf[0].strand, leftoverBasepairAllowance)  )
                    E.debug("Exon intervals preparation function: The potential half exon: "+str(exon) )
                    exon_chopped = (exon[0], exon[0]+leftoverBasepairAllowance)
                    exons_chopped.append( exon_chopped )    # append is the correct way, same as above
                    exon_choppedSize = exon_chopped[1]-exon_chopped[0]
                    leftoverBasepairAllowance -= exon_choppedSize
                    E.debug( "Exon intervals preparation function: gene %s:%s, leftoverBasepairAllowance= %s " % (gtf[0].transcript_id, gtf[0].strand, leftoverBasepairAllowance)  )
                    E.debug( str(exons_chopped) )
                    break
                    # gene is longer than my window, thus chopped, 
                    # once in this section of code, exons_chopped is the final chopped gene, thus break from loop
                    # and later, exons_chopped is being return at the end of the function.

        elif gtf[0].strand == "+":
            for exon in exons[::-1]:
                exonSize = exon[1]-exon[0]
                if leftoverBasepairAllowance > exonSize:
                    leftoverBasepairAllowance -= exonSize
                    exons_chopped.insert(0, exon)#prepend is the correct way, coz we process the last exon first. so subsequent exons shall be prepended so that the basepair ordering is the same as exon ordering.
                    E.debug( str(exons_chopped) )
                else:
                    E.debug( "Exon intervals preparation function: gene %s:%s, leftoverBasepairAllowance= %s " % (gtf[0].transcript_id, gtf[0].strand, leftoverBasepairAllowance)  )
                    E.debug( "Exon intervals preparation function: The potential half exon: "+str(exon) )
                    exon_chopped = (exon[1]-leftoverBasepairAllowance, exon[1])
                    exons_chopped.insert(0,exon_chopped)#prepend is the correct way, same as above
                    exon_choppedSize = exon_chopped[1]-exon_chopped[0]
                    leftoverBasepairAllowance -= exon_choppedSize
                    E.debug( "Exon intervals preparation function: gene %s:%s, leftoverBasepairAllowance= %s " % (gtf[0].transcript_id, gtf[0].strand, leftoverBasepairAllowance)  )
                    E.debug( str(exons_chopped) )
                    break
                    # gene is longer than my window, thus chopped, 
                    # once in this section of code, exons_chopped is the final chopped gene, thus break from loop
                    # and later, exons_chopped is being return at the end of the function.
        else:
            E.error( "In exon intervals preparation function: strand is neither + nor - in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id) )
            # in this situation, gene direction column in the GTF file have some problems, 
            # exons_chopped is still [], and is being return 
            # at the end of the function. 
            # so that his componet is skipped by the counter
            
        #gene is nicely chopped, so that no alloance is left.
        if leftoverBasepairAllowance==0: 
            return exons_chopped
            
        # gene is shorter than my window, thus it is not chopped to any shorter at all.
        # you can return three different signals to the caller function, and they have different meanings:
        # if you return the gene, it will be expanded and counted, thus might not be your desired behaviour.
        # if you return None, it will be throwing an exception in count()
        # if you return [], thus len([])==0, the whole gene will be quitely omitted. 
        elif leftoverBasepairAllowance>0: 
            E.debug( "In exon intervals preparation function: gene %s:%s is shorter than the defined threeprimebiaswindow: %s ." % (gtf[0].gene_id, gtf[0].transcript_id, extension_threePrimeBiasNotZoomed) )            
            return []
        
        #leftoverBasepairAllowance is neither ==0 nor >0, there must be a bug in the code.
        else: 
            E.error( "In exon intervals preparation function: leftoverBasepairAllowance after chopping is neither ==0 nor >0 in gene %s:%s. It must be a bug" % (gtf[0].gene_id, gtf[0].transcript_id) )
            E.error( "In exon intervals preparation function: details:: before: s%, after: s%, strand: s%."%(str(exons), str(exons_chopped), gtf[0].strand)    )  
            
        #the __chopgenestoshowthreePrimeBiasNotZoomed subfunction ends here # this return below is never be used according to the logic above. i.e. unless it enters the else: warn due to a bug, this return statement is never executed.
        return
        
            
    def count( self, gtf ):
        '''build ranges to be analyzed from a gene model.

        Returns a tuple with ranges for upstream, exons, introns, downstream.
        '''

        contig = gtf[0].contig 
        exons = GTF.asRanges( gtf, "exon" )
        # skip genes without exons
        if len(exons) == 0: return 0

        introns = Intervals.complement(exons) #Tim 31th Aug 2013
        
        if len(exons) == 0:
            E.warn( "In counter function: seems like a seriouse problem, there are no exons in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            return 0
            
        if len(introns) == 0:
            E.debug( "In counter function: no introns in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            # Tim 31th Aug 2013 :
            # do not warn here because there could be many single exon genes.
            # return 0 here because if no intron, we do not want to count this whole gene.
            # otherwise, the relative height among intron, exon, upstream, downstream will be wrong
            return 0
            
        exon_start, exon_end = exons[0][0], exons[-1][1]
        if self.scale_flanks > 0:
            self.extension_downstream = (exon_end - exon_start)*self.scale_flanks
            self.extension_upstream = (exon_end - exon_start)*self.scale_flanks
            E.debug("In counter function: scale flanks")

        if gtf[0].strand == "-":
            downstream = [ ( max(0, exon_start - self.extension_downstream), exon_start ), ] 
            upstream = [ ( exon_end, exon_end + self.extension_upstream ), ]
        else:
            upstream = [ ( max(0, exon_start - self.extension_upstream), exon_start ), ] 
            downstream = [ ( exon_end, exon_end + self.extension_downstream ), ]
        
        # within the __chopAllTranscriptToAFixedLengthForTheThreePrimeBiasCounting function, the flipping for genes on "-" strand is already taken care of. Tim 31th Aug 2013
        exons_NotZoomedAndChopped_ForThreePrimeBias   = self.__chopAllTranscriptToAFixedLengthForTheThreePrimeBiasCounting( gtf, exons,   self.extension_exons_absolute_distance_topolya   )
        introns_NotZoomedAndChopped_ForThreePrimeBias = self.__chopAllTranscriptToAFixedLengthForTheThreePrimeBiasCounting( gtf, introns, self.extension_introns_absolute_distance_topolya )
        if len(exons_NotZoomedAndChopped_ForThreePrimeBias) == 0:
            E.debug( "In counter function: after exons concat together, it is not long enough in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            # Tim 31th Aug 2013 :
            # do not warn here because there are many genes's total exon length might be short.
            # return 0 here because if no exons, we do not want to count this whole gene.
            # otherwise, the relative height of intron and exon will be wrong
            return 0        
        if len(introns_NotZoomedAndChopped_ForThreePrimeBias) == 0:
            E.debug( "In counter function: after introns concat together, it is not long enough in gene %s:%s" % (gtf[0].gene_id, gtf[0].transcript_id))
            # Tim 31th Aug 2013  :
            # do not warn here because there are many genes's total exon length might be short.
            # return 0 here because if no intron, we do not want to count this whole gene.
            # otherwise, the relative height of intron and exon will be wrong
            return 0
                        
            
        E.debug("counting upstream" )
        self.counts_upstream = self.counter.getCounts( contig, upstream, self.resolution_upstream ) 
        E.debug("counting downstream" )
        self.counts_downstream = self.counter.getCounts( contig, downstream, self.resolution_downstream )
        E.debug("counting exons NotZoomedAndChopped ForThreePrimeBias" )
        self.counts_exons_NotZoomedAndChopped_ForThreePrimeBias = self.counter.getCounts( contig, exons_NotZoomedAndChopped_ForThreePrimeBias, self.resolution_exons_absolute_distance_topolya )
        E.debug("counting introns NotZoomedAndChopped ForThreePrimeBias" )
        self.counts_introns_NotZoomedAndChopped_ForThreePrimeBias = self.counter.getCounts( contig, introns_NotZoomedAndChopped_ForThreePrimeBias, self.resolution_introns_absolute_distance_topolya )
        E.debug("counting finished" )

        
        if gtf[0].strand == "+":
            # nothings needs to be flipped for positive strand gene 
            pass 
        elif gtf[0].strand == "-":
            # flip for negative strand genes
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]
            #Tim 31th Aug 2013 
            self.counts_exons_NotZoomedAndChopped_ForThreePrimeBias   = self.counts_exons_NotZoomedAndChopped_ForThreePrimeBias[::-1]
            self.counts_introns_NotZoomedAndChopped_ForThreePrimeBias = self.counts_introns_NotZoomedAndChopped_ForThreePrimeBias[::-1]
          
        self.addLengths( upstream, 
                         exons_NotZoomedAndChopped_ForThreePrimeBias, 
                         introns_NotZoomedAndChopped_ForThreePrimeBias, 
                         downstream )                   #Tim 31th Aug 2013 
                
        self.aggregate( self.counts_upstream,
                        self.counts_exons_NotZoomedAndChopped_ForThreePrimeBias,
                        self.counts_introns_NotZoomedAndChopped_ForThreePrimeBias,                        
                        self.counts_downstream )        #Tim 31th Aug 2013 

        return 1

class UTRCounter( IntervalsCounter ):
    '''counts reads in 3'UTR and 5'UTR in addition
    to the extension outside.'''

    name = "utrprofile"
    
    def __init__(self, counter, 
                 int resolution_upstream,
                 int resolution_upstream_utr,
                 int resolution_cds,
                 int resolution_downstream_utr,
                 int resolution_downstream,
                 int extension_upstream = 0, 
                 int extension_downstream = 0,
                 *args,
                 **kwargs ):

        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream 
        self.resolution_cds = resolution_cds
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream
        self.resolution_upstream_utr = resolution_upstream_utr
        self.resolution_downstream_utr = resolution_downstream_utr

        for field, length in zip( 
            ("upstream", "upstream_utr", "cds", "downstream_utr", "downstream"),
            (resolution_upstream,
             resolution_upstream_utr,
             resolution_cds,
             resolution_downstream_utr,
             resolution_downstream ) ):
            self.add( field, length )
        
    def count( self, gtf ):
        '''build ranges to be analyzed from a gene model.

        Returns a tuple with ranges for cds, upstream_utr,
        downstream_utr, upstream, downstream.
        '''

        contig = gtf[0].contig 
        exons = GTF.asRanges( gtf, "exon" )
        # skip genes without exons
        if len(exons) == 0: return 0
        exon_start, exon_end = exons[0][0], exons[-1][1]
        self.cds = GTF.asRanges( gtf, "CDS" )
        # skip genes without CDS
        if len(self.cds) == 0: return 0

        cds_start, cds_end = self.cds[0][0], self.cds[-1][1]
        utrs = Intervals.truncate( exons, self.cds )

        if gtf[0].strand == "-":
            self.upstream_utr = [ x for x in utrs if x[0] >= cds_end ]
            self.downstream_utr = [ x for x in utrs if x[1] <= cds_start ]
            self.downstream = [ ( max(0, exon_start - self.extension_downstream), exon_start ), ] 
            self.upstream = [ ( exon_end, exon_end + self.extension_upstream ), ]
        else:
            self.upstream_utr = [ x for x in utrs if x[1] <= cds_start ]
            self.downstream_utr = [ x for x in utrs if x[0] >= cds_end ]
            self.upstream = [ ( max(0, exon_start - self.extension_upstream), exon_start ), ] 
            self.downstream = [ ( exon_end, exon_end + self.extension_downstream ), ]
        
        E.debug("counting cds" )
        self.counts_cds = self.counter.getCounts( contig, self.cds, self.resolution_cds  )
        E.debug("counting upstream" )
        self.counts_upstream = self.counter.getCounts( contig, self.upstream, self.resolution_upstream ) 
        E.debug("counting downstream" )
        self.counts_downstream = self.counter.getCounts( contig, self.downstream, self.resolution_downstream )

        if self.upstream_utr:
            E.debug("counting upstream_utr" )
            self.counts_upstream_utr = self.counter.getCounts( contig, self.upstream_utr, self.resolution_upstream_utr )
        else:
            # add pseudocounts for those genes without a UTR.
            self.counts_upstream_utr = numpy.zeros( self.resolution_upstream_utr )
            self.counts_upstream_utr += (self.counts_upstream[-1] + self.counts_cds[0]) // 2

        if self.downstream_utr:
            E.debug("counting downstream_utr" )
            self.counts_downstream_utr = self.counter.getCounts( contig, self.downstream_utr, self.resolution_downstream_utr )
        else:
            # add a pseudocount for those genes without a UTR.
            self.counts_downstream_utr = numpy.zeros( self.resolution_downstream_utr )
            self.counts_downstream_utr += (self.counts_downstream[0] + self.counts_cds[-1]) // 2

        E.debug("counting finished" )

        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_cds = self.counts_cds[::-1]
            self.counts_upstream_utr = self.counts_upstream_utr[::-1]
            self.counts_downstream_utr = self.counts_downstream_utr[::-1]
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]

        self.aggregate( self.counts_upstream,
                        self.counts_upstream_utr,
                        self.counts_cds,
                        self.counts_downstream_utr,
                        self.counts_downstream )

        return 1

class RegionCounter( GeneCounter ):
    '''count in complete region given by gene/transcript.'''

    name = "intervalprofile"
    
    def __init__(self, 
                 *args,
                 **kwargs ):

        GeneCounter.__init__(self, *args, **kwargs )

        # substitute field name 'exons' with 'interval'
        assert self.fields[1] == "exons"
        self.fields[1] = "interval"

    def count( self, gtf ):
        '''count intervals.'''

        g = GTF.Entry()
        g.copy( gtf[0] )
        g.start = gtf[0].start
        g.end = gtf[-1].end
        
        return GeneCounter.count( self, [ g ] )

class MidpointCounter( GeneCounter ):
    '''count in complete region given by gene/transcript.'''

    name = "midpointprofile"
    
    def __init__(self, counter, 
                 int resolution_upstream,
                 int resolution_downstream,
                 int extension_upstream = 0, 
                 int extension_downstream = 0,
                 int scale_flanks = 0,
                 *args,
                 **kwargs ):

        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_upstream = extension_upstream
        self.extension_downstream = extension_downstream 
        self.resolution_upstream = resolution_upstream
        self.resolution_downstream = resolution_downstream

        for field, length in zip( 
            ("upstream", "downstream"),
            (resolution_upstream,
             resolution_downstream ) ):
            self.add( field, length )
        
    def count( self, gtf ):
        '''build ranges to be analyzed from a gene model.

        Returns a tuple with ranges for exons, upstream, downstream.
        '''

        contig = gtf[0].contig 
        exons = GTF.asRanges( gtf, "exon" )
        # skip genes without exons
        if len(exons) == 0: return 0
        exon_start, exon_end = exons[0][0], exons[-1][1]
        midpoint = (exon_end - exon_start) // 2 + exon_start

        if gtf[0].strand == "-":
            downstream = [ ( max(0, midpoint - self.extension_downstream), midpoint ), ] 
            upstream = [ ( midpoint, midpoint + self.extension_upstream ), ]
        else:
            upstream = [ ( max(0, midpoint - self.extension_upstream), midpoint ), ] 
            downstream = [ ( midpoint, midpoint + self.extension_downstream ), ]
        
        E.debug("counting upstream" )
        self.counts_upstream = self.counter.getCounts( contig, upstream, self.resolution_upstream ) 
        E.debug("counting downstream" )
        self.counts_downstream = self.counter.getCounts( contig, downstream, self.resolution_downstream )

        E.debug("counting finished" )

        ## revert for negative strand
        if gtf[0].strand == "-":
            self.counts_upstream = self.counts_upstream[::-1]
            self.counts_downstream = self.counts_downstream[::-1]
            
        self.addLengths( upstream, downstream )

        self.aggregate( self.counts_upstream,
                        self.counts_downstream )

        return 1
        
class TSSCounter( IntervalsCounter ):
    '''count profile at transcription start/end site.

    The transcription start site is the beginning of the first exon.
    The transcription termination site is the end of the last exon.

    '''

    name = "tssprofile"
    
    def __init__(self, counter, extension_out = 0, extension_in = 0, *args, **kwargs):
        IntervalsCounter.__init__(self, counter, *args, **kwargs )

        self.extension_out = extension_out
        self.extension_in = extension_in 

        for field, length in zip( ("tss", "tts"),
                                  ( extension_out + extension_in,
                                    extension_out + extension_in ) ):
            self.add( field, length )
        
    def count( self, gtf ):

        contig, strand = gtf[0].contig, gtf[0].strand
        exons = GTF.asRanges( gtf, "exon" )
        # skip genes without exons
        if len(exons) == 0: return 0

        self.tss, self.tts = exons[0][0], exons[-1][1]

        # no max(0, ...) here as these ranges need to have always the same length
        if strand == "-":
            self.tts_ranges = [ (max(0, self.tss - self.extension_out), 
                                 self.tss + self.extension_in), ]
            self.tss_ranges = [ (max(0, self.tts - self.extension_in), 
                                 self.tts + self.extension_out), ]
        else:
            self.tss_ranges = [ (max(0, self.tss - self.extension_out), 
                                 self.tss + self.extension_in), ]
            self.tts_ranges = [ (max(0, self.tts - self.extension_in), 
                                 self.tts + self.extension_out), ]
        E.debug( "tss=%s, tts=%s" % (self.tss_ranges, self.tts_ranges) )

        self.counts_tss = self.counter.getCounts( contig, self.tss_ranges )
        self.counts_tts = self.counter.getCounts( contig, self.tts_ranges )

         ## revert for negative strand
        if strand == "-":
            self.counts_tss = self.counts_tss[::-1]
            self.counts_tts = self.counts_tts[::-1]
            
        self.aggregate( self.counts_tss, self.counts_tts )

        return 1

def countFromGTF(counters,
                 gtf_iterator):
    '''compute counts using counters for
    transcripts in gtf_iterator.
    '''

    c = E.Counter()
    counts = [0] * len(counters)

    E.info("starting counting" )
    names = []
    for iteration, gtf in enumerate(gtf_iterator):
        name = gtf[0].transcript_id
        E.debug( "processing %s" % (name))
        names.append(names)
        gtf.sort( key = lambda x: x.start )
        c.input += 1
        for x, counter in enumerate(counters):
            counter.update( gtf )
            counts[x] += 1

        if iteration % 100 == 0:
            E.debug( "iteration %i: counts=%s" % (iteration, ",".join( map( str, counters) ) ))

    E.info( "counts: %s: %s" % (str(c), ",".join( map(str,counts)))) 
    return names


def countFromCounts( counters,
                     all_counts ):
    '''collect counts from dataframe all_counts
    for counters.
    '''

    c = E.Counter()
    counts = [0] * len(counters)

    E.info("starting counting" )

    for iteration, x in enumerate(all_counts.iterrows()):
        c.input += 1
        name, read_counts = x
        for x, counter in enumerate(counters):
            counter.aggregate( numpy.array(read_counts) )
            counts[x] += 1

        if iteration % 100 == 0:
            E.debug( "iteration %i: counts=%s" % (iteration, ",".join( map( str, counters) ) ))

    E.info( "counts per counter: %s: %s" % (str(c), ",".join( map(str,counts)))) 
    return

        
    
        



'''gff2bed.py - convert from gff/gtf to bed
=========================================

:Tags: Genomics Intervals GFF BED Conversion

Purpose
--------

This script converts GFF or GTF formatted files to BED formatted
files.

Documentation
--------------

Users can select the field from the GTF file to be used in the name
field of the BED file using ``--set-name``. Choices include "gene_id",
"transcript_id", "class", "family", "feature", "source", "repName"
and "gene_biotype".
To specify the input is in GTF format use --is-gtf.

BED files can contain multiple tracks. If required, users can use the
"feature" or "source" fields in the input GFF file to specifiy
different tracks in the BED file (default none).

Usage
------

Example::

   # View input GTF file
   head tests/gff2bed.py/mm9_ens67_geneset_100.gtf

   # Convert GTF to bed format using gene_id as name and group by GTF feature
   cat tests/gff2bed.py/mm9_ens67_geneset_100.gtf | cgat gff2bed.py --is-gtf --set-name=gene_id --track=feature > mm9_ens67_geneset_100_feature.bed

+-------------------------------------------------------+
|track name=CDS                                         |
+------+---------+---------+--------------------+---+---+
|chr18 |3122494  |3123412  |ENSMUSG00000091539  |0  |-  |
+------+---------+---------+--------------------+---+---+
|chr18 |3327491  |3327535  |ENSMUSG00000063889  |0  |-  |
+------+---------+---------+--------------------+---+---+
|chr18 |3325358  |3325476  |ENSMUSG00000063889  |0  |-  |
+------+---------+---------+--------------------+---+---+

Command line options
--------------------

'''

import sys
import itertools
import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.Bed as Bed


def transcript2bed12(transcript):

    new_entry = Bed.Bed()
    start = min(entry.start for entry in transcript)
    end = max(entry.end for entry in transcript)

    try:
        thickStart = min(entry.start for entry in transcript 
                         if entry.feature == "CDS")
        thickEnd = max(entry.end for entry in transcript
                       if entry.feature == "CDS")
    except ValueError:

        # if there is no CDS, then set first base of transcript as 
        # start

        if transcript[0].strand == "-":
            thickStart = end
            thickEnd = end
        else:
            thickStart = start
            thickEnd = start

    exons = GTF.asRanges(transcript, "exon")

    exon_starts = [es - start for (es, ee) in exons]
    exon_lengths = [ee - es for (es, ee) in exons]
    exon_count = len(exons)
    new_entry.contig = transcript[0].contig
    new_entry.start = start
    new_entry.end = end
    new_entry["strand"] = transcript[0].strand
    new_entry["name"] = transcript[0].transcript_id

    new_entry["thickStart"] = thickStart
    new_entry["thickEnd"] = thickEnd
    
    new_entry["blockCount"] = exon_count
    new_entry["blockStarts"] = ",".join(map(str, exon_starts))
    new_entry["blockSizes"] = ",".join(map(str, exon_lengths))

    return new_entry


def main(argv=sys.argv):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("--is-gtf", dest="is_gtf", action="store_true",
                      help="input file is in gtf format [default=%default] ")

    parser.add_option(
        "--set-name", dest="name", type="choice",
        help="field from the GFF/GTF file to use as the "
        "name field in the BED file [%default]",
        choices=("gene_id", "transcript_id", "class", "family",
                 "feature", "source", "repName", "gene_biotype"))

    parser.add_option(
        "--track", dest="track", type="choice",
        choices=("feature", "source", None),
        help="use feature/source field to define BED tracks "
        "[default=%default]")

    parser.add_option(
        "--bed12-from-transcripts", dest="bed12", action="store_true",
        default=False,
        help="Process GTF file into Bed12 entries, with blocks as exons"
             "and thick/thin as coding/non-coding")

    parser.set_defaults(
        track=None,
        name="gene_id",
        is_gtf=False)

    (options, args) = E.Start(parser, add_pipe_options=True)

    ninput, noutput = 0, 0

    iterator = GTF.iterator(options.stdin)

    if options.bed12:
        iterator = GTF.transcript_iterator(iterator)

    if options.track:
        all_input = list(iterator)

        if options.track == "feature":
            grouper = lambda x: x.feature
        elif options.track == "source":
            grouper = lambda x: x.source

        all_input.sort(key=grouper)

        bed = Bed.Bed()
        for key, vals in itertools.groupby(all_input, grouper):
            options.stdout.write("track name=%s\n" % key)
            for gff in vals:
                ninput += 1
                
                if options.bed12:
                    bed = transcript2bed12(gff)
                else:
                    bed.fromGTF(gff, name=options.name)
                
                options.stdout.write(str(bed) + "\n")
                noutput += 1

    else:
        bed = Bed.Bed()
        for gff in iterator:
            ninput += 1

            if options.bed12:
                bed = transcript2bed12(gff)
            else:
                bed.fromGTF(gff, name=options.name)
            
            options.stdout.write(str(bed) + "\n")

            noutput += 1

    E.info("ninput=%i, noutput=%i" % (ninput, noutput))
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))




'''
WrapperIDR.py - wrap idr analysis
=================================

:Tags: Python

Purpose
-------

Wrap the IDR analysis workflow.

For more information, see

https://sites.google.com/site/anshulkundaje/projects/idr


Usage
-----

Documentation
-------------

Code
----

This script requires some R functions within the package
called 

functions-all-clayton-12-13.r

These functions have been copied into a script called
WrapperIDR.r in the scripts directory.

'''

import os
import sys

from CGAT import Experiment as E

from rpy2.robjects import r as R


def runIDR(options, peakfile1, peakfile2):
    '''run IDR analysis.

    This code is taken from the R script

    batch-consistency-analysis.r
    '''

    if options.half_width is not None:
        R.assign("half.width", options.half_width)
    else:
        R('''half.width = NULL''')
    R.assign("overlap.ratio", options.overlap_ratio)
    R.assign("is.broadpeak", options.is_broadpeak)
    R.assign("sig.value", options.signal_value)

    dirname = os.path.dirname(__file__)
    R.source(os.path.join(dirname, "WrapperIDR.r"))

    # read the length of the chromosomes, which will be used to concatenate
    # chr's
    R('''chr.size <- read.table('%s', sep='\t')''' %
      options.filename_chromosome_table)

    output_prefix = options.output_prefix
    output_uri = output_prefix + "-uri.sav"
    output_em = output_prefix + "-em.sav"
    output_overlapped_peaks = output_prefix + "-overlapped-peaks.txt"
    output_peaks_above_idr = output_prefix + "-npeaks-aboveIDR.txt"

    # process data, summit: the representation of the location of summit
    E.info("loading data")
    R('''rep1 <- process.narrowpeak('%(peakfile1)s', chr.size, 
                     half.width=half.width, summit="offset", broadpeak=is.broadpeak)''' % locals())
    R('''rep2 <- process.narrowpeak('%(peakfile2)s', chr.size, 
                     half.width=half.width, summit="offset", broadpeak=is.broadpeak)''' % locals())

    E.info("replicate 1: read %s: %i peaks, %i after filtering" %
           (peakfile1,
            R('''nrow(rep1$data.ori)''')[0],
            R('''nrow(rep1$data.cleaned)''')[0]))
    E.info("replicate 2: read %s: %i peaks, %i after filtering" %
           (peakfile2,
            R('''nrow(rep2$data.ori)''')[0],
            R('''nrow(rep2$data.cleaned)''')[0]))

    E.info("computing correspondence profile (URI)")

    R('''uri.output <- compute.pair.uri(rep1$data.cleaned, rep2$data.cleaned, 
                                        sig.value1=sig.value, sig.value2=sig.value, 
                                        overlap.ratio=overlap.ratio)''')
    E.info("saving correspondence profile to %s" % output_uri)
    R('''save(uri.output, file='%(output_uri)s') ''' % locals())

    E.info("computing EM procedure for inference")
    R('''em.output <- fit.em(uri.output$data12.enrich, fix.rho2=T)''')
    E.info("saving EM to %s" % output_em)
    R('''save(em.output, file='%(output_em)s') ''' % locals())

    # write em output into a file
    # cat(paste("EM estimation for the following files\n", peakfile1, "\n", peakfile2, "\n", sep=""))

    options.stdout.write("em_estimation\n%s\n" %
                         str(R('''em.output$em.fit$para''')))

    # add on 3-29-10
    # output both local idr and IDR
    E.info("writing overlapped peaks to %s" % output_overlapped_peaks)
    R('''idr.local <- 1-em.output$em.fit$e.z''')
    R('''IDR <- c()''')
    R('''o <- order(idr.local)''')
    R('''IDR[o] <- cumsum(idr.local[o])/c(1:length(o))''')
    R('''
    write.out.data <- data.frame(chr1=em.output$data.pruned$sample1[, "chr"],
                                 start1=em.output$data.pruned$sample1[, "start.ori"],
                                 stop1=em.output$data.pruned$sample1[, "stop.ori"],
                                 sig.value1=em.output$data.pruned$sample1[, "sig.value"],
                                 chr2=em.output$data.pruned$sample2[, "chr"],
                                 start2=em.output$data.pruned$sample2[, "start.ori"],
                                 stop2=em.output$data.pruned$sample2[, "stop.ori"],
                                 sig.value2=em.output$data.pruned$sample2[, "sig.value"],
                                 idr.local=1-em.output$em.fit$e.z, IDR=IDR)
    ''')
    R('''write.table(write.out.data, file='%(output_overlapped_peaks)s')''' %
      locals())

    # number of peaks passing IDR range (0.01-0.25)
    E.info("computing number of peaks at various thresholds")
    R('''IDR.cutoff <- seq(0.01, 0.25, by=0.01)''')
    R('''idr.o <- order(write.out.data$idr.local)''')
    R('''idr.ordered <- write.out.data$idr.local[idr.o]''')
    R('''IDR.sum <- cumsum(idr.ordered)/c(1:length(idr.ordered))''')
    R('''
    IDR.count <- c()
    n.cutoff <- length(IDR.cutoff)
    for(i in 1:n.cutoff){
        IDR.count[i] <- sum(IDR.sum <= IDR.cutoff[i])
        }
    ''')

    # write the number of peaks passing various IDR ranges into a file
    E.info(
        "writing number of peaks above IDR cutoffs in range [0.01, 0.25] to %s" % output_peaks_above_idr)
    R('''idr.cut <- data.frame( cutoff=IDR.cutoff, count=IDR.count)''')
    R('''write.table(idr.cut, file='%(output_peaks_above_idr)s', quote=F, 
                     row.names=F, col.names=T, sep='\t')''' % locals())

    R('''mar.mean <- get.mar.mean(em.output$em.fit)''')
    options.stdout.write("marginal mean of two components\n%s\n)" %
                         R('''print(mar.mean)'''))


def plotIDR(output_file, input_prefixes):
    '''create IDR plots.

    This code is taken from the R script

    batch-consistency-plot.r

    within the IDR package.
    '''

    dirname = os.path.dirname(__file__)
    R.source(os.path.join(dirname, "WrapperIDR.r"))

    R('''df.txt = 10''')

    R('''uri.list <- list()
         uri.list.match <- list()
         ez.list <- list()
         legend.txt <- c()
         em.output.list <- list()
         uri.output.list <- list()''')

    npair = len(input_prefixes)
    for x, input_prefix in enumerate(input_prefixes):

        R.load(input_prefix + "-uri.sav")
        R.load(input_prefix + "-em.sav")
        i = x + 1

        R( '''uri.output.list[[%(i)i]] <- uri.output;
              em.output.list[[%(i)i]] <- em.output;
              # reverse =T for error rate;''' % locals())
        R('''
              ez.list[[%(i)i]] <- get.ez.tt.all(em.output, uri.output.list[[%(i)i]]$data12.enrich$merge1,
                                        uri.output.list[[%(i)i]]$data12.enrich$merge2);''' % locals())
        R('''
              # URI for all peaks
              uri.list[[%(i)i]] <- uri.output$uri.n;

              # URI for matched peaks
              uri.match <- get.uri.matched(em.output$data.pruned, df=df.txt);
              uri.list.match[[%(i)i]] <- uri.match$uri.n;
         ''' % locals() )

        legend = "%(i)i = %(input_prefix)s" % locals()
        R('''
              legend.txt[%(i)i] <- '%(legend)s';
        ''' % locals())

    R.pdf(output_file)
    R('''par(mfcol=c(2,3), mar=c(5,6,4,2)+0.1)''')
    R('''plot.uri.group(uri.list, NULL, file.name=NULL, c(1:%(npair)i), title.txt="all peaks");
         plot.uri.group(uri.list.match, NULL, file.name=NULL, c(1:%(npair)i), title.txt="matched peaks");
         plot.ez.group(ez.list, plot.dir=NULL, file.name=NULL, legend.txt=c(1:%(npair)i), y.lim=c(0, 0.6));
         plot(0, 1, type="n", xlim=c(0,1), ylim=c(0,1), xlab="", ylab="", xaxt="n", yaxt="n"); 
         legend(0, 1, legend.txt, cex=0.6);''' % locals())
    R["dev.off"]()


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $",
                            usage=globals()["__doc__"])

    parser.add_option("-o", "--output-prefix", dest="output_prefix", type="string",
                      help="output filename prefix [default=%default].")

    parser.add_option("-c", "--chromosome-table", dest="filename_chromosome_table", type="string",
                      help="filename with tab separated list of chromosome names [default=%default].")

    parser.add_option("--action", dest="action", type="choice",
                      choices=("plot", "run"),
                      help="action to perform [default=%default]")

    parser.add_option("-s", "--signal-value", dest="signal_value", type="string",
                      help="use either p.value or sig.value as ranking measure [default=%default]")

    parser.add_option("-r", "--overlap-ratio", dest="overlap_ratio", type="int",
                      help="a value between 0 and 1 that controls how much two peaks have to overlap to be called as the same [default=%default]")

    parser.set_defaults(
        action="plot",
        output_prefix="output",
        half_width=None,
        overlap_ratio=0,
        is_broadpeak=False,
        signal_value="signal.value",
        filename_chromosome_table="genome_table.txt",
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if options.action == "plot":
        plotIDR(options.output_prefix + ".pdf", args)
    elif options.action == "run":
        if len(args) != 2:
            raise ValueError("require exactly two replicates")
        runIDR(options, args[0], args[1])

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''Intervals.py - Utility functions for working with intervals
===========================================================

This module contains utility functions for working intervals
or list of intervals.

An interval is a tuple of a start and end coordinate in python's
0-based, half-open notation such as::

    (12, 20)

An interval list is simply a list of such intervals.

The majority of the functions in this module take one or more lists of
intervals and return one or more new lists of intervals.

Reference
---------

'''


def getLength(intervals):
    """return sum of intervals lengths.

    >>> getLength([(10,20), (30,40)])
    20

    """
    return sum([x[1] - x[0] for x in intervals])


def combine(intervals):
    """combine overlapping and adjacent intervals.

    >>> combine([(10,20), (30,40)])
    [(10, 20), (30, 40)]
    >>> combine([(10,20), (20,40)])
    [(10, 40)]
    >>> combine([(10,20), (15,40)])
    [(10, 40)]
    """
    if not intervals:
        return []

    new_intervals = []

    intervals.sort()
    first_from, last_to = intervals[0]

    for this_from, this_to in intervals[1:]:
        if this_from > last_to:
            new_intervals.append((first_from, last_to))
            first_from, last_to = this_from, this_to
            continue

        if last_to < this_to:
            last_to = this_to

    new_intervals.append((first_from, last_to))

    return new_intervals


def prune(intervals, first=None, last=None):
    """truncates all intervals that are extending beyond first or last.

    Empty intervals are deleted.
    """
    new_intervals = []
    for start, end in intervals:
        if end <= first or start >= end:
            continue
        new_intervals.append((max(first, start), min(end, last)))
    return new_intervals


def complement(intervals, first=None, last=None):
    """complement a list of intervals with intervals not in list.

    >>> complement([(10,20), (15,40)])
    []
    >>> complement([(10,20), (30,40)])
    [(20, 30)]
    >>> complement([(10,20), (30,40)], first=5)
    [(5, 10), (20, 30)]

    Arguments
    ---------
    intervals : list
        List of intervals
    first : int
        First position. If given, the interval from `first` to
        the first position in `intervals` is added.
    last : int
        Last position. If given, the interval from the last position
        in `intervals` to `last` is added.

    Returns
    -------
    intervals : list
        A new list of intervals
    """

    if len(intervals) == 0:
        if first is not None and last is not None:
            return [(first, last)]
        else:
            return []

    new_intervals = []

    intervals.sort()
    last_from, last_to = intervals[0]

    if first is not None and first < last_from:
        new_intervals.append((first, last_from))

    for this_from, this_to in intervals:
        if this_from > last_to:
            new_intervals.append((last_to, this_from))

        last_from = this_from
        last_to = max(last_to, this_to)

    if last and last > last_to:
        new_intervals.append((last_to, last))

    return new_intervals


def addComplementIntervals(intervals, first=None, last=None):
    """complement a list of intervals with intervals not
    in list and return both.

    The resulting interval list is sorted.
    """
    return sorted(intervals + complement(intervals, first, last))


def joined_iterator(intervals1, intervals2):
    """iterate over the combination of two intervals.

    returns the truncated intervals delineating the
    ranges of overlap between intervals1 and intervals2.
    """

    intervals1.sort()
    intervals2.sort()

    x, y = 0, 0
    while x < len(intervals1) and y < len(intervals2):

        xfrom, xto = intervals1[x]
        yfrom, yto = intervals2[y]

        if xto <= yfrom:
            x += 1
        elif yto <= xfrom:
            y += 1
        else:
            mto = min(xto, yto)
            mfrom = max(xfrom, yfrom)

            yield (mfrom, mto)

            if xto < yto:
                x += 1
            elif yto < xto:
                y += 1
            else:
                x += 1
                y += 1


def intersect(intervals1, intervals2):
    """intersect two interval sets.

    Return a set of intervals that is spanned by intervals in
    both sets. Returns the union of the two intervals.
    """
    if not intervals1 or not intervals2:
        return []

    result = []
    for start, end in joined_iterator(intervals1, intervals2):

        overlap = end - start
        if overlap:
            result.append((start, end))

    return result


def truncate(intervals1, intervals2):
    """truncate intervals in intervals1 by intervals2

    Example: truncate( [(0,5)], [(0,3)] ) = [(3,5)]
    """
    if not intervals1:
        return []
    if not intervals2:
        return intervals1

    intervals2 = combine(intervals2)

    result = []
    intersection = intersect(intervals1, intervals2)

    if not intersection:
        return intervals1

    y = 0
    for start, end in intervals1:

        # retrieve all overlapping in intersection
        ovl = []
        while y < len(intersection):
            mask_start, mask_end = intersection[y]
            if mask_start >= end:
                break
            if mask_start >= start:
                ovl.append((mask_start, mask_end))
            y += 1

        # build truncated segments
        last = start
        for mask_start, mask_end in ovl:
            if mask_start - last > 0:
                result.append((last, mask_start))
            last = mask_end
        if end - last > 0:
            result.append((last, end))

    return result


def calculateOverlap(intervals1, intervals2):
    """calculate overlap between two list of intervals.

    The intervals within each set should not be overlapping.
    """

    if not intervals1 or not intervals2:
        return 0

    overlap = 0
    for start, end in joined_iterator(intervals1, intervals2):
        overlap += end - start

    return overlap


def fromArray(a):
    """get intervals from a binary array."""

    if len(a) == 0:
        return []

    was_a = a[0]
    if was_a:
        start = 0

    intervals = []
    for x in range(1, len(a)):
        is_a = a[x]
        if is_a and not was_a:
            start = x
            was_a = is_a
        elif not is_a and was_a:
            if start is not None:
                intervals.append((start, x))
            was_a = is_a
    if is_a:
        intervals.append((start, len(a)))

    return intervals


def combineAtDistance(intervals, min_distance):
    """combine a list intervals and merge those that are less than a
    certain distance apart.
    """

    if not intervals:
        return []

    # merge overlapping intervals
    intervals = combine(intervals)

    new_intervals = []
    intervals.sort()

    first_start, last_end = intervals[0]

    for this_start, this_end in intervals[1:]:

        if this_start - last_end >= min_distance:
            new_intervals.append((first_start, last_end))
            first_start = this_start

        last_end = this_end

    new_intervals.append((first_start, last_end))

    return new_intervals


def getIntersections(intervals):
    """return regions were two intervals are overlapping.

    """
    if not intervals:
        return []

    intervals.sort()
    max_to = intervals[0][1]

    all_sections = []
    sections = [intervals[0][0], intervals[0][1]]
    for this_from, this_to in intervals[1:]:

        # no overlap: write everything and reset
        if this_from > max_to:
            all_sections.append(sections)
            max_to = this_to
            sections = []

        max_to = max(max_to, this_to)

        sections.append(this_from)
        sections.append(this_to)

    all_sections.append(sections)

    new_intervals = []
    for sections in all_sections:
        sections.sort()
        last_x = sections[0]
        for x in sections[1:]:
            if last_x == x:
                continue
            new_intervals.append((last_x, x))
            last_x = x

    return new_intervals


def RemoveIntervalsContained(intervals):
    """remove intervals that are fully contained in another

    [(10, 100), (20, 50), (70, 120), (130, 200), (10, 50), (140, 210), (150, 200)]

    results:

    [(10, 100), (70, 120), (130, 200), (140, 210)]
    """
    if not intervals:
        return []

    new_intervals = []

    intervals.sort()
    last_from, last_to = intervals[0]

    for this_from, this_to in intervals[1:]:
        # this is larger:
        if this_from <= last_from and this_to >= last_to:
            last_from, last_to = this_from, this_to
            continue

        # last is larger
        if last_from <= this_from and last_to >= this_to:
            continue

        # no complete overlap
        new_intervals.append((last_from, last_to))

        last_from, last_to = this_from, this_to

    new_intervals.append((last_from, last_to))

    return new_intervals


def RemoveIntervalsSpanning(intervals):
    """remove intervals that are full covering
    another, i.e. always keep the smallest.

    [(10, 100), (20, 50), (70, 120), (40,80), (130, 200), (10, 50), (140, 210), (150, 200)]

    result:

    [(20, 50), (40, 80), (70, 120), (150, 200)]
    """

    if not intervals:
        return []

    intervals.sort()

    last_intervals = intervals

    while 1:

        new_intervals = []

        last_from, last_to = last_intervals[0]

        for this_from, this_to in last_intervals[1:]:
            # print last_from, last_to, this_from, this_to
            # this is larger:
            if this_from <= last_from and this_to >= last_to:
                continue

            # last is larger:
            if last_from <= this_from and last_to >= this_to:
                last_from, last_to = this_from, this_to
                continue

            # no complete overlap
            new_intervals.append((last_from, last_to))
            last_from, last_to = this_from, this_to

        new_intervals.append((last_from, last_to))

        if len(last_intervals) == len(new_intervals):
            break

        last_intervals = new_intervals

    return new_intervals


def ShortenIntervalsOverlap(intervals, to_remove):
    """shorten intervals, so that there is no
    overlap with another set of intervals.

    assumption: intervals are not overlapping

    """
    if not intervals:
        return []

    if not to_remove:
        return interalls

    new_intervals = []

    intervals.sort()
    to_remove.sort()

    current_to_remove = 0

    for this_from, this_to in intervals:

        for remove_from, remove_to in to_remove:
            # print this_from, this_to, remove_from, remove_to
            if remove_to < this_from:
                continue
            if remove_from > this_to:
                continue

            if remove_from <= this_from and remove_to >= this_to:
                this_from = remove_to
                break

            if this_from < remove_from:
                new_intervals.append((this_from, remove_from))
                # print "adding", this_from, remove_from

            this_from = max(this_from, remove_to)

            if this_to < this_from:
                break

        if this_to > this_from:
            # print "adding", this_from, this_to
            new_intervals.append((this_from, this_to))

    return new_intervals



'''Fastq.py - methods for dealing with fastq files
===============================================

This module provides an iterator of :term:`fastq` formatted files
(:func:`iterate`). Additional iterators allow guessing of the quality
score format (:func:`iterate_guess`) or converting them
(:func:`iterate_convert`) while iterating through a file.

:func:`guessFormat` inspects a fastq file to guess the quality score format
and :func:`getOffset` returns the numeric offset for quality score conversion
for a particular quality score format.

.. note::
   Another way to access the information in :term:`fastq` formatted
   files is through pysam_.

Reference
---------

'''

import string

from math import log

import CGAT.Experiment as E
import CGAT.IOTools as IOTools

RANGES = {
    'sanger': (33, 75),
    'illumina-1.8': (33, 79),
    'solexa': (59, 106),
    'phred64': (64, 106),
}


class Record:
    """A record representing a :term:`fastq` formatted record.

    Attributes
    ----------
    identifier : string
       Sequence identifier
    seq : string
       Sequence
    quals : string
       String representation of quality scores.
    format : string
       Quality score format. Can be one of ``sanger``,
       ``illumina-1.8``, ``solexa`` or ``phred64``.

    """

    def __init__(self, identifier, seq, quals, format=None):
        self.identifier, self.seq, self.quals, format = (
            identifier, seq, quals, format)
        self.format = None

    def __str__(self):
        return "@%s\n%s\n+\n%s" % (self.identifier, self.seq, self.quals)

    def guessFormat(self):
        '''return quality score format -
        might return several if ambiguous.'''

        c = [ord(x) for x in self.quals]
        mi, ma = min(c), max(c)
        r = []
        for format, v in RANGES.items():
            m1, m2 = v
            if mi >= m1 and ma <= m2:
                r.append(format)
        return r

    def guessDataType(self):
        '''return the datatype. This is done by inspecting the
        sequence for basecalls/colorspace ints'''

        datatypes = set()

        # don't check the first character as it may be a basecall even with cs
        if sum([x in ["0", "1", "2", "3", "."]
                for x in self.seq[1:]]) == len(self.seq[1:]):
            datatypes.add("colorspace")
        elif sum([x.upper() in string.ascii_uppercase
                  for x in self.seq]) == len(self.seq):
            datatypes.add("basecalls")

        return datatypes

    def trim(self, trim3, trim5=0):
        """remove nucleotides/quality scores from the 3' and 5' ends."""
        self.seq = self.seq[trim5:-trim3]
        self.quals = self.quals[trim5:-trim3]

    def trim5(self, trim5=0):
        """remove nucleotides/quality scores from the 5' ends."""
        self.seq = self.seq[trim5:]
        self.quals = self.quals[trim5:]

    def toPhred(self):
        '''return qualities as a list of phred-scores.'''
        assert self.format is not None, "format needs to be set for conversion"
        if self.format == "sanger":
            return [ord(x) - 33 for x in self.quals]
        elif self.format == "illumina-1.8":
            return [ord(x) - 33 for x in self.quals]
        elif self.format == "solexa":
            # from -5 to 40 (i.e., can be negative)
            log10x = log(10.0) + .499
            return [int(10.0 * log(1.0 + 10 ** (ord(x) / 10.0), 10) / log10x)
                    for x in self.quals]
        elif self.format == "phred64":
            return [ord(x) - 64 for x in self.quals]

    def fromPhred(self, quals, format):
        '''set qualities from a list of phred-scores.'''
        self.format = format
        # -1 for color space fastq file
        assert len(quals) == len(self.seq) or len(quals) == len(self.seq) - 1
        if self.format == "sanger":
            self.quals = "".join([chr(33 + x) for x in quals])
        elif self.format == "illumina-1.8":
            return [chr(33 + x) for x in quals]
        elif self.format == "solexa":
            log10x = log(10.0, 10) / 10.0
            q = [int(10.0 * (log(10 ** (x * log10x) - 1.0, 10)))
                 for x in quals]
            self.quals = "".join([chr(64 + x) for x in q])
        elif self.format == "phred64":
            self.quals = "".join([chr(64 + x) for x in quals])
        elif self.format == "integer":
            self.quals = " ".join(map(str, quals))


def iterate(infile):
    '''iterate over contents of fastq file.'''

    while 1:
        line1 = infile.readline()
        if not line1:
            break
        if not line1.startswith('@'):
            raise ValueError("parsing error: expected '@' in line %s" % line1)
        line2 = infile.readline()
        line3 = infile.readline()
        if not line3.startswith('+'):
            raise ValueError("parsing error: expected '+' in line %s" % line3)
        line4 = infile.readline()
        # incomplete entry
        if not line4:
            raise ValueError("incomplete entry for %s" % line1)

        yield Record(line1[1:-1], line2[:-1], line4[:-1])


def iterate_guess(infile, max_tries=10000, guess=None):
    '''iterate over contents of fastq file.

    Guess quality format by looking at the first `max_tries` entries and
    then subsequently setting the quality score format for each entry.

    Arguments
    ---------
    infile : File
       File or file-like object to iterate over
    max_tries : int
       Number of records to examine for guessing the quality score
       format.
    guess : string
       Default format. This format will be chosen in the quality
       score format is ambiguous. The method checks if the `guess`
       is compatible with the records read so far.

    Yields
    ------
    fastq
        An object of type :class:`Record`.

    Raises
    ------
    ValueError
        If the ranges of the fastq records are not compatible,
        are incompatible with guess or are ambiguous.

    '''
    quals = set(RANGES.keys())
    cache = []
    myiter = iterate(infile)
    lengths = []
    for c, record in enumerate(myiter):
        quals.intersection_update(set(record.guessFormat()))
        if len(quals) == 0:
            raise ValueError("could not guess format - ranges incompatible.")
        if len(quals) == 1:
            break
        cache.append(record)
        lengths.append(len(record.seq))
        if c > max_tries:
            break

    if len(quals) == 1:
        ref_format = list(quals)[0]
    elif guess in quals:
        E.warn("multiple input formats possible: %s. Continuing with %s" %
               (", ".join(quals), guess))
        ref_format = guess
    elif quals.issubset(set(["solexa", "phred64"])):
        # guessFormat will call phred64 reads as phred64 AND solexa
        # if both still remain after max_tries, assume phred64
        ref_format = "phred64"
    else:
        raise ValueError(
            "could not guess format - could be one of %s." % str(quals))

    for r in cache:
        r.format = ref_format
        yield r

    for r in myiter:
        r.format = ref_format
        yield r


def iterate_convert(infile, format, max_tries=10000, guess=None):
    '''iterate over contents of fastq file.

    The quality score format is guessed and all subsequent records
    are converted to `format`.

    Arguments
    ---------
    infile : File
       File or file-like object to iterate over
    format : string
       Quality score format to convert all records into.
    max_tries : int
       Number of records to examine for guessing the quality score
       format.
    guess : string
       Default format. This format will be chosen in the quality
       score format is ambiguous. The method checks if the `guess`
       is compatible with the records read so far.

    Yields
    ------
    fastq
        An object of type :class:`Record`.

    Raises
    ------
    ValueError
        If the ranges of the fastq records are not compatible,
        are incompatible with guess or are ambiguous.


    '''

    quals = set(RANGES.keys())
    cache = []
    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        quals.intersection_update(set(record.guessFormat()))

        if len(quals) == 0:
            raise ValueError("could not guess format - ranges incompatible.")
        if len(quals) == 1:
            cache.append(record)
            break
        cache.append(record)

        if c > max_tries:
            break

    if len(quals) == 1:
        ref_format = list(quals)[0]
    elif quals.issubset(set(["solexa", "phred64"])):
        # guessFormat will call phred64 reads as phred64 AND solexa
        # if both still remain after max_tries, assume phred64
        ref_format = "phred64"
    elif guess in quals:
        E.warn("multiple input formats possible: %s. Continuing with %s" %
               (", ".join(quals), guess))
        ref_format = guess
    else:
        raise ValueError(
            "could not guess format - could be one of %s. "
            "If you know the format use the --format option" % str(quals))

    for r in cache:
        r.format = ref_format
        r.fromPhred(r.toPhred(), format)
        yield r

    for r in myiter:
        r.format = ref_format
        r.fromPhred(r.toPhred(), format)
        yield r


def guessFormat(infile, max_lines=10000, raises=True):
    '''guess format of FASTQ File.

    Arguments
    ---------
    infile : File
       File or file-like object to iterate over
    max_lines : int
       Number of lines to examine for guessing the quality score
       format.
    raises : bool
       Raise ValueError if format is ambiguous

    Returns
    -------
    formats : list
       list of quality score formats compatible with the file

    Raises
    ------
    ValueError
        If the ranges of the fastq records are not compatible.

    '''

    quals = set(RANGES.keys())
    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        quals.intersection_update(set(record.guessFormat()))
        if len(quals) == 0:
            raise ValueError("could not guess format - ranges incompatible.")
        if len(quals) == 1:
            break
        if c > max_lines:
            break

    if len(quals) == 1:
        return list(quals)[0]
    elif raises is False:
        if quals.issubset(set(["solexa", "phred64"])):
            # guessFormat will call phred64 reads as phred64 AND solexa
            # if both still remain after max_lines, return phred64
            return "phred64"
        else:
            return quals
    else:
        raise ValueError(
            "could not guess format - could be one of %s." % str(quals))


def guessDataType(infile, max_lines=10000, raises=True):
    '''guess datatype of FASTQ File from [colourspace, basecalls]

    Arguments
    ---------
    infile : File
    File or file-like object to iterate over
    max_lines : int
       Number of lines to examine for guessing the datatype
    raises : bool
       Raise ValueError if format is ambiguous

    Returns
    -------
    formats : list
       list of datatypes compatible with the file (should only ever be one!)

    Raises
    ------
    ValueError
        If the ranges of the fastq records are not compatible.

    '''

    datatype = set(("colorspace", "basecalls"))

    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        datatype.intersection_update(record.guessDataType())

        if len(datatype) == 0:
            raise ValueError("could not guess datatype")
        if len(datatype) == 1:
            break
        if c > max_lines:
            break

    if len(datatype) == 1:
        return list(datatype)[0]
    else:
        raise ValueError(
            "could not guess datatype - could be one of %s." % str(datatype))


def getOffset(format, raises=True):
    '''returns the ASCII offset for a certain format.

    If `raises` is set a ValueError is raised if there is not a single
    offset. Otherwise, a minimum offset is returned.

    Returns
    -------
    offset : int
       The quality score offset

    '''
    if type(format) in (set, list, tuple):
        offsets = set([RANGES[f][0] for f in format])
        if len(offsets) == 1:
            return list(offsets)[0]
        elif raises is False:
            return min(offsets)
        else:
            raise ValueError(
                "inconsistent offsets for multiple formats: %s" % offsets)

    return RANGES[format][0]


def getReadLength(filename):
    '''return readlength from a fastq file.

    Only the first read is inspected. If there are
    different read lengths in the file, the result
    will be inaccurate.

    Returns
    -------
    read_length : int
    '''

    with IOTools.openFile(filename) as infile:
        record = next(iterate(infile))
        return len(record.seq)



'''fasta2table.py - analyze sequence composition, codon usage, bias and degeneracy 
====================================================================================

:Tags: Genomics Sequences

Purpose
-------

This script reads a collection of sequences in :term:`fasta` format, computes
various sequence properties and outputs them to a tab-separated file.

Documentation
--------------

The following sequence properties can be reported:

length
   sequence length and number of codons (0 for amino-acid sequence)

sequence
   Full nucleotide / amino-acid sequence

hid
   a hash identifier for the sequence

na
   nucleic acid composition, including GC/AT content

dn
   dinucleotide counts

cpg
   CpG density and CpG observed/expected

gaps
    number of gaps and gapped/ungapped regions in the sequences

aa
   amino acid composition (nucleotide sequence must have length divisible by 3)

degeneracy
    count the number of degenerate sites (nucleotide sequence only,
    sequence must have length divisible by 3)

codons
   codon composition (nucleotide sequence only, sequence must have
   length divisible by 3)

codon-usage
    output codon frequencies for each sequence (nucleotide sequence
    only, sequence must have length divisible by 3)

codon-bias
    output codon bias for each sequence (nucleotide sequence only,
    sequence must have length divisible by 3)

codon-translator
    translate codons for each sequence to their frequency (nucleotide
    sequence only, sequence must have length divisible by 3)

Multiple counters can be calculated at the same by specifying 
--section multiple times.

The script can also process fasta description lines (starting >)
either by splitting each line at the first space and taking only the
first part (--split-fasta-identifier), or by any user-supplied python
regular expression (--regex-identifier).

Usage
-----

Example::

   # View fasta file
   head tests/fasta2table.py/na_test.fasta

   # Count CpG dinucleotides
   cgat fasta2table --sections=length,na --split-fasta-identifier < tests/fasta2table.py/test.fasta > na.tsv

In this example we input a fasta file and compute the sequence composition, i.e.
%C, %G, %A, %T as well for each sequence in the set.

+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|id              |length       |nC     |nG     |nA     |nT     |nN   |nUnk |nGC    |nAT    |nCpG  |pC        |pG        |pA        |pT        |pN        |pUnk      |pGC       |pAT       |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-0        |110529       |26348  |22351  |34036  |27794  |0    |0    |48699  |61830  |4990  |0.238381  |0.202218  |0.307937  |0.251463  |0.000000  |0.000000  |0.440599  |0.559401  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-1000000  |121046       |28432  |22851  |36540  |33223  |0    |0    |51283  |69763  |4814  |0.234886  |0.188779  |0.301869  |0.274466  |0.000000  |0.000000  |0.423665  |0.576335  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-2000000  |61046        |11368  |14668  |16666  |18344  |0    |0    |26036  |35010  |2526  |0.186220  |0.240278  |0.273007  |0.300495  |0.000000  |0.000000  |0.426498  |0.573502  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-3000000  |76120        |17653  |14496  |23591  |20380  |0    |0    |32149  |43971  |3099  |0.231910  |0.190436  |0.309919  |0.267735  |0.000000  |0.000000  |0.422346  |0.577654  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-4000000  |66656        |12970  |14994  |19076  |19616  |0    |0    |27964  |38692  |2911  |0.194581  |0.224946  |0.286186  |0.294287  |0.000000  |0.000000  |0.419527  |0.580473  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-5000000  |81994        |15755  |19061  |23070  |24108  |0    |0    |34816  |47178  |3571  |0.192148  |0.232468  |0.281362  |0.294022  |0.000000  |0.000000  |0.424616  |0.575384  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+
|contig-6000000  |106529       |16997  |21653  |21112  |22258  |0    |0    |38650  |43370  |5135  |0.207230  |0.263997  |0.257401  |0.271373  |0.000000  |0.000000  |0.471227  |0.528773  |
+----------------+-------------+-------+-------+-------+-------+-----+-----+-------+-------+------+----------+----------+----------+----------+----------+----------+----------+----------+

Type::

   cgat fasta2table --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import math

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.IOTools as IOTools
import CGAT.SequenceProperties as SequenceProperties
import CGAT.FastaIterator as FastaIterator


def main(argv=None):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-w", "--weights-tsv-file", dest="filename_weights",
        type="string",
        help="filename with codon frequencies. Multiple filenames "
        "can be separated by comma.")

    parser.add_option(
        "-s", "--section", dest="sections", type="choice", action="append",
        choices=("length", "sequence", "hid", "na", "aa", "cpg", "dn",
                 "degeneracy", "gaps",
                 "codons", "codon-usage", "codon-translator", "codon-bias"),
        help="which sections to output [%default]")

    parser.add_option(
        "-t", "--sequence-type", dest="seqtype", type="choice",
        choices=("na", "aa"),
        help="type of sequence: na=nucleotides, aa=amino acids [%default].")

    parser.add_option(
        "-e", "--regex-identifier", dest="regex_identifier", type="string",
        help="regular expression to extract identifier from fasta "
        "description line.")

    parser.add_option(
        "--split-fasta-identifier", dest="split_id",
        action="store_true",
        help="split fasta description line (starting >) and use "
        "only text before first space")

    parser.add_option(
        "--add-total", dest="add_total", action="store_true",
        help="add a row with column totals at the end of the table"
        "[%default]")

    parser.set_defaults(
        filename_weights=None,
        pseudocounts=1,
        sections=[],
        regex_identifier="(.+)",
        seqtype="na",
        gap_chars='xXnN',
        split_id=False,
        add_total=False,
    )

    (options, args) = E.Start(parser, argv=argv)

    rx = re.compile(options.regex_identifier)

    reference_codons = []
    if options.filename_weights:
        options.filename_weights = options.filename_weights.split(",")
        for filename in options.filename_weights:
            if filename == "uniform":
                reference_codons.append(Genomics.GetUniformCodonUsage())
            else:
                reference_codons.append(
                    IOTools.ReadMap(IOTools.openFile(filename, "r"),
                                    has_header=True,
                                    map_functions=(str, float)))

        # print codon table differences
        options.stdlog.write(
            "# Difference between supplied codon usage preferences.\n")
        for x in range(0, len(reference_codons)):
            for y in range(0, len(reference_codons)):
                if x == y:
                    continue
                # calculate KL distance
                a = reference_codons[x]
                b = reference_codons[y]
                d = 0
                for codon, p in list(a.items()):
                    if Genomics.IsStopCodon(codon):
                        continue
                    d += b[codon] * math.log(b[codon] / p)

                options.stdlog.write("# tablediff\t%s\t%s\t%f\n" %
                                     (options.filename_weights[x],
                                      options.filename_weights[y],
                                      d))

    iterator = FastaIterator.FastaIterator(options.stdin)

    def getCounter(section):

        if options.seqtype == "na":
            if section == "length":
                s = SequenceProperties.SequencePropertiesLength()
            elif section == "sequence":
                s = SequenceProperties.SequencePropertiesSequence()
            elif section == "hid":
                s = SequenceProperties.SequencePropertiesHid()
            elif section == "na":
                s = SequenceProperties.SequencePropertiesNA()
            elif section == "gaps":
                s = SequenceProperties.SequencePropertiesGaps(
                    options.gap_chars)
            elif section == "cpg":
                s = SequenceProperties.SequencePropertiesCpg()
            elif section == "dn":
                s = SequenceProperties.SequencePropertiesDN()
            # these sections requires sequence length to be a multiple of 3
            elif section == "aa":
                s = SequenceProperties.SequencePropertiesAA()
            elif section == "degeneracy":
                s = SequenceProperties.SequencePropertiesDegeneracy()
            elif section == "codon-bias":
                s = SequenceProperties.SequencePropertiesBias(reference_codons)
            elif section == "codons":
                s = SequenceProperties.SequencePropertiesCodons()
            elif section == "codon-usage":
                s = SequenceProperties.SequencePropertiesCodonUsage()
            elif section == "codon-translator":
                s = SequenceProperties.SequencePropertiesCodonTranslator()
            else:
                raise ValueError("unknown section %s" % section)
        elif options.seqtype == "aa":
            if section == "length":
                s = SequenceProperties.SequencePropertiesLength()
            elif section == "sequence":
                s = SequenceProperties.SequencePropertiesSequence()
            elif section == "hid":
                s = SequenceProperties.SequencePropertiesHid()
            elif section == "aa":
                s = SequenceProperties.SequencePropertiesAminoAcids()
            else:
                raise ValueError("unknown section %s" % section)
        return s

    # setup totals
    totals = {}
    for section in options.sections:
        totals[section] = getCounter(section)

    options.stdout.write("id")
    for section in options.sections:
        options.stdout.write("\t" + "\t".join(totals[section].getHeaders()))

    options.stdout.write("\n")
    options.stdout.flush()

    s = getCounter("hid")
    s.loadSequence("AAAAAAAAA", "na")

    for cur_record in iterator:

        sequence = re.sub(" ", "", cur_record.sequence).upper()

        if len(sequence) == 0:
            raise ValueError("empty sequence %s" % cur_record.title)

        id = rx.search(cur_record.title).groups()[0]

        if options.split_id is True:
            options.stdout.write("%s" % id.split()[0])
        else:
            options.stdout.write("%s" % id)
        options.stdout.flush()

        for section in options.sections:
            s = getCounter(section)
            s.loadSequence(sequence, options.seqtype)
            totals[section].addProperties(s)

            options.stdout.write("\t" + "\t".join(s.getFields()))

        options.stdout.write("\n")

    if options.add_total:
        options.stdout.write("total")
        for section in options.sections:
            options.stdout.write("\t" + "\t".join(totals[section].getFields()))
        options.stdout.write("\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""bam2bed.py - convert bam formatted file to bed formatted file
================================================================

:Tags: Genomics NGS Intervals BAM BED Conversion

Purpose
-------

This tool converts BAM files into BED files supplying the intervals
for each read in the BAM file.  BAM files must have a corresponding
index file ie. example.bam and example.bam.bai

For example::

   samtools view example.bam

   READ1    163    1      13040   15     76M    =      13183   219     ...
   READ1    83     1      13183   7      76M    =      13040   -219    ...
   READ2    147    1      13207   0      76M    =      13120   -163    ...

   python bam2bed.py example.bam

   1       13039   13115   READ1     15      +
   1       13119   13195   READ2     0       +
   1       13182   13258   READ1     7       -
   1       13206   13282   READ2     0       -

By default, bam2bed outputs each read as a separate interval.  With
the option ``--merge-pairs`` paired-end reads are merged and output as
a single interval. The strand is set according to the first read in a
pair.

Usage
-----

::

   cgat bam2bed BAMFILE [--merge-pairs] [options]

operates on the file BAMFILE::

   cgat bam2bed [--merge-pairs] [options]

operates on the stdin as does::

   cgat bam2bed -I BAMFILE [--merge-pairs] [options]


To merge paired-end reads and output fragment interval ie. leftmost
mapped base to rightmost mapped base::

   cat example.bam | cgat bam2bed --merge-pairs

   1       13119   13282   READ2     0       +
   1       13039   13258   READ1     7       +

To use merge pairs on only a region of the genome use samtools view::

   samtools view -ub example.bam 1:13000:13100 | cgat bam2bed --merge-pairs

Note that this will select fragments were the first read-in-pair is in
the region.

Options
-------

-m, --merge-pairs
    Output one region per fragment rather than one region per read,
    thus a single region is create stretching from the start of the
    frist read in pair to the end of the second.

    Read pairs that meet the following criteria are removed:

    * Reads where one of the pair is unmapped
    * Reads that are not paired
    * Reads where the pairs are mapped to different chromosomes
    * Reads where the the insert size is not between the max and
      min (see below)

.. warning::

    Merged fragements are always returned on the +ve strand.
    Fragement end point is estimated as the alignment start position
    of the second-in-pair read + the length of the first-in-pair
    read. This may lead to inaccuracy if you have an intron-aware
    aligner.

--max-insert-size, --min-insert-size
    The maximum and minimum size of the insert that is allowed when
    using the --merge-pairs option. Read pairs closer to gether or futher
    apart than the min and max repsectively are skipped.

-b, --bed-format
    What format to output the results in. The first n columns of the bed
    file will be output.



Type::

   python bam2bed.py --help

for command line help.

Command line options
--------------------

"""

import sys
import pysam
import CGAT.Experiment as E
import CGAT.scripts._bam2bed as _bam2bed


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id$", usage=globals()["__doc__"])

    parser.add_option("-m", "--merge-pairs", dest="merge_pairs",
                      action="store_true",
                      help="merge paired-ended reads and output interval "
                      "for entire fragment [default=%default]. ")

    parser.add_option("--max-insert-size", dest="max_insert_size", type="int",
                      help="only merge paired-end reads if they are less than "
                      "# bases apart. "
                      " 0 turns off this filter [default=%default]. ")

    parser.add_option("--min-insert-size", dest="min_insert_size", type="int",
                      help="only merge paired-end reads if they are at "
                      "least # bases apart. "
                      " 0 turns off this filter [default=%default]. ")

    parser.add_option("--bed-format", dest="bed_format", type="choice",
                      choices=('3', '4', '5', '6'),
                      help="bed format to output. "
                      " [default=%default]")

    parser.set_defaults(
        region=None,
        call_peaks=None,
        merge_pairs=None,
        min_insert_size=0,
        max_insert_size=0,
        bed_format='6',
    )

    (options, args) = E.Start(parser, argv=argv)

    if len(args) == 0:
        args.append("-")

    samfile = pysam.AlignmentFile(args[0], "rb")

    options.bed_format = int(options.bed_format)

    if options.merge_pairs is not None:
        counter = _bam2bed.merge_pairs(samfile,
                                       options.stdout,
                                       min_insert_size=options.min_insert_size,
                                       max_insert_size=options.max_insert_size,
                                       bed_format=options.bed_format)

        E.info("category\tcounts\n%s\n" % counter.asTable())

    else:
        # use until_eof. Files from stdin have no index
        it = samfile.fetch(until_eof=True)

        # more comfortable cigar parsing will
        # come with the next pysam release
        BAM_CMATCH = 0
        BAM_CDEL = 2
        BAM_CREF_SKIP = 3
        take = (BAM_CMATCH, BAM_CDEL, BAM_CREF_SKIP)
        outfile = options.stdout

        for read in it:
            if read.is_unmapped:
                continue

            t = 0
            for op, l in read.cigar:
                if op in take:
                    t += l

            if read.is_reverse:
                strand = "-"
            else:
                strand = "+"
            # IMS: converted rname to reference name
            outfile.write("%s\t%d\t%d\t%s\t%d\t%c\n" %
                          (samfile.getrname(read.rname),
                           read.pos,
                           read.pos + t,
                           read.qname,
                           read.mapq,
                           strand))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''fasta2variants.py - create sequence variants from a set of sequences
====================================================================

:Tags: Genomics Sequences Variants Protein FASTA Transformation

Purpose
-------

This script reads a collection of sequences in :term:`fasta` format
and outputs a table of possible variants. It outputs for each position
in a protein sequence the number of variants.

If the input sequences are nucleotide coding (CDS) sequences, for each
variant a weight is output indicating the number of times that variant
can occur from single nucleotide changes.

Usage
-----

Example::

    python fasta2variants.py -I CCDS_nucleotide.current.fna.gz -L CDS.log -S CDS.output -c

This will take a CDS file as input, save the log and output files, and
count variants based on single nucleotide changes using the -c option.

Type::

    python fasta2variants.py --help

for command line help.

Compressed (.gz) and various fasta format files (.fasta, .fna) are
accepted. If the -c option is specified and the file is not a CDS
sequence the script will throw an error ('length of sequence
'<input_file>' is not a multiple of 3').

Command line options
--------------------

'''
import sys
import collections

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.FastaIterator as FastaIterator


def main(argv=None):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-c", "--is-cds", dest="is_cds", action="store_true",
                      help="input are cds (nucleotide) sequences [%default]")

    parser.set_defaults(
        is_cds=False,
    )

    (options, args) = E.Start(parser, argv=argv)

    options.stdout.write(
        "snpid\tidentifier\tpos\treference\tvariant\tcounts\tweight\n")

    alphabet = "ACDEFGHIKLMNPQRSTVWY"

    snpid = 0

    for entry in FastaIterator.iterate(options.stdin):
        identifier = entry.title

        if options.is_cds:
            cds_sequence = entry.sequence.upper()
            assert len(cds_sequence) % 3 == 0, \
                "length of sequence '%s' is not a multiple of 3" % entry.title

            sequence = Genomics.translate(cds_sequence)
            weights = []
            for pos, cds_pos in enumerate(range(0, len(cds_sequence), 3)):
                codon = cds_sequence[cds_pos:cds_pos + 3]
                counts = collections.defaultdict(int)
                for x in range(0, 3):
                    rna = codon[x]
                    for na in "ACGT":
                        if na == rna:
                            continue
                        taa = Genomics.translate(
                            codon[:x] + na + codon[x + 1:])
                        counts[taa] += 1
                weights.append(counts)

        else:
            sequence = entry.sequence.upper()
            counts = {}
            for x in alphabet:
                counts[x] = 1
            weights = [counts] * len(sequence)

        for pos, ref in enumerate(sequence):

            if ref not in alphabet:
                continue
            w = weights[pos]
            t = float(sum(w.values()))
            for variant in alphabet:
                if variant == ref:
                    continue
                snpid += 1
                options.stdout.write(
                    "%s\n" % "\t".join(
                        ("%010i" % snpid,
                         identifier,
                         str(pos + 1),
                         ref,
                         variant,
                         "%i" % w[variant],
                         "%6.4f" % (w[variant] / t),
                         )))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
VCF.py - Tools for working with VCF files
=========================================

The parser for VCF files is very simplistic.

.. note::
   Another way to access the information in :term:`vcf` formatted
   files is through pysam_.

The Variant Call Format (:term:`vcf`) is described
at http://www.1000genomes.org/wiki/doku.php?id=1000_genomes:analysis:vcf4.0

Reference
---------

'''
import sys


class VCFEntry:
    """A VCF Entry"""

    def __init__(self, data, samples):

        assert len(data) == len(samples) + 9
        self.contig, self.pos, self.id, self.ref, self.alt, self.qual, \
            self.filter, self.info, self.format = \
            data[:9]

        self.genotypes = dict(list(zip(samples, data[9:])))
        self.order = samples

    def __str__(self):
        return "\t".join(map(str, (
            self.contig, self.pos, self.id, self.ref, self.alt, self.qual,
            self.filter, self.info, self.format,
            "\t".join([self.genotypes[x] for x in self.order]))))


class VCFFile:
    """A VCF File"""

    def __init__(self, infile):

        self.infile = infile
        self.format = {}
        self.info = {}
        self.fileformat = None

        while 1:
            line = self.infile.readline()

            if line.startswith("##"):
                self.addMetaFromLine(line[2:-1])
                continue
            elif line.startswith("#CHROM"):
                self.samples = line[:-1].split("\t")[9:]
                continue
            elif line.startswith("#"):
                continue
            break
        self.line = line

    def writeHeader(self, outfile, order=None):
        outfile.write("##fileformat=%s\n" % self.fileformat)
        for key, values in self.format.items():
            outfile.write("##FORMAT=%s,%s\n" % (key, ",".join(values)))
        for key, values in self.info.items():
            outfile.write("##INFO=%s,%s\n" % (key, ",".join(values)))
        outfile.write(
            "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t")
        if order:
            assert len(order) == len(self.samples), \
                "number of samples do not match: %i != %i" % (
                    len(order), len(self.samples))
            outfile.write("\t".join(order))
        else:
            outfile.write("\t".join(self.samples))
        outfile.write("\n")

    def __iter__(self):
        return self

    def addMetaFromLine(self, line):

        key, value = line.split("=", 1)
        if key == "INFO":
            data = value.split(",")
            self.info[data[0]] = data[1:]
        elif key == "FORMAT":
            data = value.split(",")
            self.format[data[0]] = data[1:]
        elif key == "fileformat":
            self.fileformat = value

    def __next__(self):

        data = self.line[:-1].split("\t")
        self.line = self.infile.readline()
        if not self.line:
            raise StopIteration
        return VCFEntry(data, self.samples)

    def next(self):
        return self.__next__()

if __name__ == "__main__":

    inf = VCFFile(sys.stdin)

    for x in inf:
        print(str(x))



'''chain2stats.py
==============

:Tags: Genomics GenomeAlignment Summary CHAIN

Purpose
-------

This script takes a UCSC :term: `chain` file and computes coverage across the
target and query genomes.

As different regions of a genome vary in the extent to which they are
conserved between species, genomic local alignments will inevitably
result in discrete regions of successfully aligned sequence separated
by regions of the genome where alignment is not possible. Chains
consist of gapless blocks of successfully aligned regions, separated
by regions where alignment is not possible. This script is returns
summary statistics that relate the proportion of the query and target
genomes that are covered by chains in a chain file.


Usage
-----

Example::

python chain2stats.py < mm10ToHg19.over.chain.gz

Expected output::

    ********** chain2stats report starts **********

    Report for chains per chromosome gapped statistics
    --------------------------------------------------
    Target genome 91.31, query genome: 76.17


    Report for chains per chromosome ungapped statistics
    ----------------------------------------------------
    Target genome 37.02, query genome: 29.13


    Report for chain lengths gapped
    -------------------------------
    number of chains = 76560

    Stats for target chain lengths:
    Mean:36,516, median:456, max:91,540,769, min:23

    Stats for query chain lengths:
    Mean:43,401, median:480, max:120,333,803, min:23


    Report for chain lengths ungapped
    ---------------------------------
    number of chains = 33902022

    Stats for target chain lengths:
    Mean:30, median:21, max:17,154, min:1

    Stats for query chain lengths:
    Mean:30, median:21, max:17,154, min:1


    ********** chain2stats report ends **********

    Type::

    python chain2stats.py --help

    for command line help.


Options
-------

By default chain2stats will report the percentage of target and query
genomes that are covered by gapped and ungapped chains as well as summary
statistics (mean, median, max, min) for gapped and ungapped chain lengths.
Additional options available are:

``--aggregate-by``
    Will cause coverage to be calculated on a per chromosome basis if
    set to ``chromosome``

``--output-identity``

    Will report summary statistics (mean, median, max, min) for the
    percent identity between gapped regions. Requires ``--dbpath``,
    ``--target-genome``, and ``--query-genome`` to be set.

``--dbpath``

    Directory containing indexed fasta files for target and query
    genome.

``--target-genome``
    Filename prefix for target genome fasta file.

``--query-genome``
    Filename prefix for query genome fasta file.

``--output-report``
    Write summary statistics for target and query genomes to separate
    tab-delimited outfiles.

``--errors``
    Report inconsitencies between chain contig sizes and contig sizes in
    the supplied fasta files. Requires ``--dbpath``, ``--target-genome``,
    and ``--query-genome`` to be set.


Command line options
---------------------

'''

import os
import sys
import bx.bitset
import collections
import numpy

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta


def chain_iterator(infile):
    lines = []
    for line in infile:
        if line.startswith("#"):
            continue
        if line.strip() == "":
            continue
        if line.startswith("chain"):
            if lines:
                yield lines
            lines = []
        lines.append(line)
    yield lines


class Chain:
    "class with methods for getting gapped and ungapped regions represented "
    "by a chain"

    def __init__(self, data):
        self.lines = data
        self.header = self.lines[0]
        self.atts = self.header.split()
        a = len(self.atts)
        if a != 13:  # sanity check, can remove
            raise Exception("Unexpected Header length\n" + self.header)
        self.score = int(self.atts[1])
        self.tname = self.atts[2]
        self.tsize = int(self.atts[3])
        self.tstrand = self.atts[4]
        self.tstart = int(self.atts[5])
        self.tend = int(self.atts[6])
        self.qname = self.atts[7]
        self.qsize = int(self.atts[8])
        self.qstrand = self.atts[9]
        b = self.qstrand
        if (b != "+") and (b != "-"):
            raise Exception("unexpected strand id")
        self.qstart = int(self.atts[10])
        self.qend = int(self.atts[11])
        self.id = int(self.atts[12])
        self.tlength = self.tend - self.tstart
        self.qlength = self.qend - self.qstart
        self.nlines = len(self.lines)
        self.nblocks = self.nlines - 1
        (self.tgr, self.qgr) = self._gapped_regions()
        (self.tugr, self.qugr) = self._ungapped_regions()

    def _gapped_regions(self):
        tr, qr = [], []
        tr.append([self.tstart, self.tlength])
        qr.append([self.qstart, self.qlength])
        return(tr, qr)

    def _ungapped_regions(self):
        ts = self.tstart
        qs = self.qstart
        tr, qr = [], []
        for i in range(1, self.nlines):
            dataline = self.lines[i]
            values = dataline.split()
            len_align = int(values[0])
            te = ts + len_align
            qe = qs + len_align
            tr.append([ts, len_align])
            qr.append([qs, len_align])
            if (i < self.nlines - 1):
                ts = te + int(values[1])
                qs = qe + int(values[2])
        return(tr, qr)

    def get_regions(self, gapped):
        if gapped is True:
            treg, qreg = self.tgr, self.qgr
        else:
            treg, qreg = self.tugr, self.qugr
        return(treg, qreg)


class Bitsets_Container:

    "General class for holding bitsets in a {key1: {key2: bitset }} structure"

    def __init__(self):
        self.bitsets = collections.defaultdict(dict)

    def addRanges(self, key1, key2, size, ranges):
        if key2 not in self.bitsets[key1]:
            if size == 199501827:  # this randomly segfaults a binnedbitset
                self.bitsets[key1][key2] = bx.bitset.BitSet(size)
            else:
                self.bitsets[key1][key2] = bx.bitset.BinnedBitSet(size)
        for r in ranges:  # bx-python does sanity checks.
            self.bitsets[key1][key2].set_range(r[0], r[1])

    def bitset_coverage(self, key1, key2):
        b = self.bitsets[key1][key2]
        l = b.size
        c = b.count_range(0, l)
        return (c, l)


class ChainCounter():

    '''Top, defining class for a chain counter.

    Counters must have:
    (1) an add method to add a new chain
    (2) a write_report method

    Counters will generally initialise a container and add metrics to
    it when passed a chain

    '''

    def add(self, c):
        raise NotImplementedError("abstract method - implement in subclass")

    def _get_stats(self, options):
        '''Calculates and stores the stats if necessary - called by all report
        methods
        '''
        raise NotImplementedError("abstract method - implement in subclass")

    def report(self, options):
        '''should call the "write_report" method to do the writing'''
        raise NotImplementedError("abstract method - implement in subclass")

    def tabbed_report(self, options):
        '''should call the "write_tabbed" method to do the writing'''
        raise NotImplementError("abstract method - implement in subclass")

    def _get_basic_stats(self, numlist, string="{:.2f}"):
        s = [string.format(numpy.mean(numlist)),
             string.format(
                 numpy.median(numlist)), string.format(max(numlist)),
             string.format(min(numlist))]
        return(s)

    def _wrap_basic_stats(self, s, tabbed=False):
        if tabbed is False:
            r = ''.join(
                ["Mean:", s[0], ", median:", s[1], ", max:", s[2],
                 ", min:", s[3], "\n"])
        else:
            r = '\t'.join(s)
        return(r)

    def _write_report(self, options, report):
        options.stdout.write(''.join(['\n', '\n'.join(report), '\n']))

    def _write_tabbed(self, name, lines, E):
        outfile = E.openOutputFile(name)
        outfile.write('\n'.join(lines))
        outfile.write('\n')
        outfile.close

    def _wrap_header(self, text=""):
        '''decorates header for report'''
        if text == "":
            text = self.header
        underline = ''.join(["-" for i in text])
        return([text, underline])


class CounterPerChromosome(ChainCounter):

    header = "Report for chains per chromosome"

    def __init__(self, gapped):
        self.container = Bitsets_Container()
        self.gapped = gapped
        self.cov = 0
        if self.gapped is True:
            self.name = "gapped"
        else:
            self.name = "ungapped"

    def _pc_coverage(self, coverage, total_length):
        pc_cov = float(coverage) / float(total_length) * 100
        return(pc_cov)

    def _coverage(self, key1):
        total, total_len = 0, 0
        d = self.container.bitsets[key1]
        for key2 in sorted(d):
            (s, l) = self.container.bitset_coverage(key1, key2)
            total += s
            total_len += l
        gen_coverage = self._pc_coverage(total, total_len)
        return gen_coverage

    def add(self, c):
        treg, qreg = c.get_regions(self.gapped)
        self.container.addRanges("target", c.tname, c.tsize, treg)
        self.container.addRanges("query", c.qname, c.qsize, qreg)

    def _get_stats(self):
        if self.cov == 0:
            self.cov = {}
            for i in ("target", "query"):
                self.cov[i] = self._coverage(i)

    def report(self, options):
        self._get_stats()
        report = self._wrap_header(
            "".join([self.header, " ", self.name, " statistics"]))
        report.append("Target genome {:.2f}, query genome: {:.2f}\n".format(
            self.cov["target"], self.cov["query"]))
        self._write_report(options, report)

    def tabbed_report(self, options, E):
        self._get_stats()
        lines = ["target_genome\tquery_genome"]
        lines.append(
            '{:.2f}\t{:.2f}'.format(self.cov["target"], self.cov["query"]))
        self._write_tabbed(''.join(["genome_coverage_", self.name]), lines, E)


class CounterPerChromosomePair(CounterPerChromosome):

    header = "Report for chains per chromosome pair"

    def _make_table(self, options):
        table = []
        table.append("t_chrom\tt_cov\tq_chrom\tq_cov\tbases aligned")
        tchrom = self.cov[0][0]
        j = 0
        for i in self.cov:
            if i[0] != tchrom:
                tchrom = i[0]
                j = 0
            if j < options.nperchrom:
                table.append(
                    "{}\t{:.2f}\t{}\t{:.2f}\t{:,}".format(i[0], i[1], i[2], i[3], i[4]))
                j += 1
            else:
                continue
        table.append("")
        return(table)

    def add(self, c):
        treg, qreg = c.get_regions(self.gapped)
        if c.qstrand == "+":
            strand = "plus"
        else:
            strand = "minus"
        pair_name = (c.tname, c.qname)
        self.container.addRanges(pair_name, "target", c.tsize, treg)
        self.container.addRanges(pair_name, "query", c.qsize, qreg)
        self.container.addRanges(pair_name, strand, c.qsize, qreg)

    def _get_stats(self, options):
        if self.cov == 0:
            result = []
            for key1 in self.container.bitsets:
                (tm, tl) = self.container.bitset_coverage(key1, 'target')
                tc = self._pc_coverage(tm, tl)
                (qm, ql) = self.container.bitset_coverage(key1, 'query')
                qc = self._pc_coverage(qm, ql)
                result.append([key1[0], tc, key1[1], qc, tm])
            a = sorted(result, key=itemgetter(4), reverse=True)
            self.cov = sorted(a, key=itemgetter(0))
            self.cov_table = self._make_table(options)

    def report(self, options):
        self._get_stats(options)
        report = (
            self._wrap_header("".join([self.header, " ", self.name, " statistics"])))
        report = report + self.cov_table
        self._write_report(options, report)

    def tabbed_report(self, options, E):
        self._get_stats(options)
        self._write_tabbed(
            ''.join(["chrom_pair_cov_", self.name]), self.cov_table, E)


class CounterOfGappedChainLengths(ChainCounter):

    header = "Report for chain lengths"

    def __init__(self, gapped):
        self.nchains = 0
        self.tcls, self.qcls = [], []
        self.stats = {}
        self.gapped = gapped
        if self.gapped is True:
            self.name = "gapped"
        else:
            self.name = "ungapped"

    def _chain_stats_report(self, name):
        report = [''.join(["Stats for ", name, " chain lengths:"])]
        report.append(self._wrap_basic_stats(self.stats[name]))
        return(report)

    def add(self, c):
        treg, qreg = c.get_regions(self.gapped)
        n = len(treg)
        if n != len(qreg):
            raise Exception(
                "Different numbers of chains for query and target!?!")
        for i in range(0, n):
            self.tcls.append(treg[i][1])
            self.qcls.append(qreg[i][1])
            self.nchains += 1

    def _get_stats(self):
        if ("target" in self.stats) == 0:
            self.stats["target"] = self._get_basic_stats(
                self.tcls, string="{:,.0f}")
        if ("query" in self.stats) == 0:
            self.stats["query"] = self._get_basic_stats(
                self.qcls, string="{:,.0f}")

    def report(self, options):
        self._get_stats()
        report = (self._wrap_header(" ".join([self.header, self.name])))
        report.append("number of chains = {:}\n".format(self.nchains))
        for i in ("target", "query"):
            report = report + self._chain_stats_report(i)
        self._write_report(options, report)

    def tabbed_report(self, options, E):
        self._get_stats()
        for i in ("target", "query"):
            lines = ["mean\tmedian\tmax\tmin"]
            lines.append(self._wrap_basic_stats(self.stats[i], tabbed=True))
            self._write_tabbed('_'.join([i, "lengths", self.name]), lines, E)


class CounterPercentIdentify(ChainCounter):

    header = "Report on Percent Indentities"

    def __init__(self, tpath, qpath):
        self.tfasta = IndexedFasta.IndexedFasta(tpath)
        self.qfasta = IndexedFasta.IndexedFasta(qpath)
        self.pids = []
        self.stats = 0

    def _get_pid(self, x, y):
        z = list(zip(x, y))
        pid = (float(len([a for a, b in z if a == b])) / float(len(z)) * 100)
        return(pid)

    def add(self, c):
        nreg = len(c.tugr)
        for i in range(0, nreg):
            tseq = self.tfasta.getSequence(
                c.tname, "+", c.tugr[i][0], (sum(c.tugr[i])))
            qseq = self.qfasta.getSequence(
                c.qname, c.qstrand, c.qugr[i][0], (sum(c.qugr[i])))
            pid = self._get_pid(tseq.lower(), qseq.lower())
        self.pids.append(pid)

    def _get_stats(self):
        if self.stats == 0:
            self.stats = self._get_basic_stats(self.pids, string="{:.2f}")

    def report(self, options):
        self._get_stats()
        report = self._wrap_header()
        report.append(self._wrap_basic_stats(self.stats))
        self._write_report(options, report)

    def tabbed_report(self, options, E):
        self._get_stats()
        lines = ["mean\tmedian\tmax\tmin"]
        lines.append(self._wrap_basic_stats(self.stats, tabbed=True))
        self._write_tabbed("pids", lines, E)


class CounterOfErrors(ChainCounter):

    '''class for reporting invalid contig sizes in chains'''

    header = "Contig size validation report"

    def __init__(self, options):
        self.tdb = IndexedFasta(options.dbpath + options.targetgenome)
        self.qdb = IndexedFasta(options.dbpath + options.querygenome)
        self.tcontigs = self.tdb.getContigSizes()
        self.qcontigs = self.qdb.getContigSizes()
        self.badchains = []

    def add(self, c):
        db_tsize = self.tcontigs[c.tname]
        db_qsize = self.qcontigs[c.qname]
        if c.tsize != db_tsize:
            self.badchains.append(
                '\t'.join([str(x) for x in c.atts] + [" #bad target contigsize"]))
        if c.qsize != db_qsize:
            self.badchains.append(
                '\t'.join([str(x) for x in c.atts] + [" #bad query contigsize"]))

    def report(self, options):
        report = self._wrap_header()
        if len(self.badchains) == 0:
            report.append("All chains passed validation")
        else:
            report = report + self.badchains
        self._write_report(options, report)

    def tabbed_report(self, options, E):
        if len(self.badchains) > 0:
            lines = self.badchains
        else:
            lines = ["#no bad chains found"]
        self._write_tabbed("bad_contig_sizes", lines, E)


def main(argv=None):

    if not argv:
        argv = sys.argv

    # get the options
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-c", "--chain-file", dest="chainfile", type="string",
                      help="the chain file to analyse", metavar="FILE")

    parser.add_option(
        "--alignments-per-contig", dest="nperchrom", type="int",
        help="Number of aligments to report on per chromosome pair", default=2)

    parser.add_option(
        "--aggregate-by", dest="aggregate", type="choice",
        choices=("contig", "none"),
        help="Set to `contig` to perform per chromosome pair analysis",
        default=2)

    parser.add_option(
        "-i", "--output-identity", dest="output_identity", action="store_true",
        help="Generate stats on the sequence identity of the gapped "
        "chains. Requires FastaIndex.py", default=False)

    parser.add_option(
        "-d", "--dbpath", dest="dbpath", type="string",
        help="The path to the indexed fasta files",
        default=".")

    parser.add_option(
        "-t", "--target-genome", dest="targetgenome", type="string",
        help="The target genome, eg. Mm19", default=False)

    parser.add_option(
        "-q", "--query-genome", dest="querygenome", type="string",
        help="The query genome eg. Hg17", default=False)

    parser.add_option(
        "-e", "--errors", dest="errors", action="store_true",
        help="Check chains for erroneous contig sizes using the given db",
        default=False)

    parser.add_option(
        "-r", "--output-report", dest="output_report", action="store_true",
        help="Write out tab-delimited reports for each analysis",
        default=False)

    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    # make a list of counting objects
    counters = []

    counters.append(CounterPerChromosome(gapped=True))
    counters.append(CounterPerChromosome(gapped=False))

    if options.aggregate == "contig":
        counters.append(CounterPerChromosomePair(gapped=True))
        counters.append(CounterPerChromosomePair(gapped=False))

    counters.append(CounterOfGappedChainLengths(gapped=True))
    counters.append(CounterOfGappedChainLengths(gapped=False))

    if options.output_identity is True:
        if options.targetgenome == 0 or options.querygenome == 0:
            raise Exception(
                "Target and query database must be specified with the \"-e\" flag")
        t_db_path = os.path.join(options.dbpath, options.targetgenome)
        q_db_path = os.path.join(options.dbpath, options.querygenome)
        counters.append(CounterPercentIdentify(t_db_path, q_db_path))

    if options.errors is True:
        if options.targetgenome == 0 or options.querygenome == 0:
            raise Exception(
                "Target and query database must be specified with the \"-e\" flag")
        counters.append(CounterOfErrors(options))

    # iterate over the chains and counters
    for chain in chain_iterator(options.stdin):
        c = Chain(chain)
        for counter in counters:
            counter.add(c)

    # write a report to stdout and individual reports to tab delimited files
    options.stdout.write(
        "\n\n********** chain2stats report starts **********\n")

    for counter in counters:
        counter.report(options)
        if options.output_report is True:
            counter.tabbed_report(options, E)

    options.stdout.write("\n********** chain2stats report ends **********\n\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''Database.py - Database utility functions
===========================================

This module contains convenience functions to work with a relational
database.


Reference
---------

'''
import time
import re
from pandas import DataFrame


def executewait(dbhandle, statement, error=Exception, regex_error="locked",
                retries=-1, wait=5):
    '''repeatedly execute an SQL statement until it succeeds.


    Arguments
    ---------
    dbhandle : object
        A DB-API conform database handle.
    statement : string
        SQL statement to execute.
    error : string
        Exception to catch and examine for error messages.
    regex_error : string
        Any error message matching `regex_error` will be ignored,
        otherwise the procedure exists.
    retries : int
        Number of retries. If set to negative number, retry indefinitely.
        If set to 0, there will be only one attempt.
    wait : int
        Number of seconds to way between retries.

    Returns
    -------
    A cursor object

    '''
    cc = dbhandle.cursor()

    while 1:
        try:
            cc.execute(statement)
        except error as msg:
            if retries == 0:
                raise
            if not re.search("locked", str(msg)):
                raise
            time.sleep(wait)
            retries -= 1
            continue
        break
    return cc


def getColumnNames(dbhandle, table):
    """return column names of a table from a database.
    """

    cc = executewait(dbhandle, "SELECT * FROM %s LIMIT 1" % table)
    return tuple([x[0] for x in cc.description])


def getTables(dbhandle):
    """get list of tables in an sqlite database"""
    cc = executewait(
        dbhandle, """select name from sqlite_master where type='table'""")
    return tuple([x[0] for x in cc])


def toTSV(dbhandle, outfile, statement, remove_none=True):
    '''execute statement and save as tsv file
    to disk.

    If *remove_none* is true, empty/NULL values will be output as
    empty values.

    '''
    cc = dbhandle.cursor()
    cc.execute(statement)
    outfile.write("\t".join([x[0] for x in cc.description]) + "\n")

    def _str(x):
        if x is None:
            return ""
        else:
            return str(x)

    if remove_none:
        f = _str
    else:
        f = str

    outfile.write("\n".join(
        ["\t".join(map(f, x)) for x in cc]))


def db_execute(cc, statements):
    '''excute a statement or statements against a cursor'''

    if type(statements) not in (list, tuple):
        statements = [statements]

    for statement in statements:
        cc.execute(statement)


def connect(dbhandle, attach=None):
    """attempt to connect to database.

    If `dbhandle` is an existing connection to a database,
    it will be returned unchanged. Otherwise, this method
    will attempt to establish a connection.

    Arguments
    ---------
    dbhandle : object or string
        A database handle or a connection string.

    Returns
    -------
    dbhandle : object
        A DB-API2 conforming database handle
    """
    if isinstance(dbhandle, str):
        try:
            import sqlite3
        except ImportError:
            raise ValueError(
                "If an sqlite database location is passed"
                " directly the sqlite3 module must be installed")

        dbhandle = sqlite3.connect(dbhandle)

    cc = dbhandle.cursor()

    if attach is not None:
        if isinstance(attach, str):
            db_execute(cc, attach)
        elif isinstance(attach, (tuple, list)):
            for attach_statement in attach:
                db_execute(cc, attach_statement)

    return dbhandle


def execute(queries, dbhandle=None, attach=False):
    '''Execute a statement or a  list of statements (sequentially)'''

    cc = dbhandle.cursor()

    if attach:
        db_execute(cc, attach)

    db_execute(cc, queries)
    cc.close()


def fetch(query, dbhandle=None, attach=False):
    '''Fetch all query results and return'''

    cc = dbhandle.cursor()

    if attach:
        db_execute(cc, attach)

    sqlresult = cc.execute(query).fetchall()
    cc.close()
    return sqlresult


def fetch_with_names(query,
                     dbhandle=None,
                     attach=False):
    '''Fetch query results and returns them as an array of row arrays, in
       which the first entry is an array of the field names

    '''

    dbhandle = connect(dbhandle, attach=attach)

    cc = dbhandle.cursor()
    sqlresult = cc.execute(query).fetchall()

    data = []
    # http://stackoverflow.com/questions/4147707/
    # python-mysqldb-sqlite-result-as-dictionary
    field_names = [d[0] for d in cc.description]
    data.append([name for name in field_names])
    for record in sqlresult:
        line = [field for field in record]
        data.append(line)

    cc.close()
    return data


def fetch_DataFrame(query,
                    dbhandle=None,
                    attach=False):
    '''Fetch query results and returns them as a pandas dataframe'''

    dbhandle = connect(dbhandle, attach=attach)

    cc = dbhandle.cursor()
    sqlresult = cc.execute(query).fetchall()
    cc.close()

    # see http://pandas.pydata.org/pandas-docs/dev/generated/
    # pandas.DataFrame.from_records.html#pandas.DataFrame.from_records
    # this method is design to handle sql_records with proper type
    # conversion

    field_names = [d[0] for d in cc.description]
    pandas_DataFrame = DataFrame.from_records(
        sqlresult,
        columns=field_names)
    return pandas_DataFrame


def write_DataFrame(dataframe,
                    tablename,
                    dbhandle=None,
                    index=False,
                    if_exists='replace'):
    '''write a pandas dataframe to an sqlite db, index on given columns
       index columns given as a string or list eg. "gene_id" or
       ["gene_id", "start"]

    '''

    dbhandle = connect(dbhandle)

    dataframe.to_sql(tablename,
                     con=dbhandle,
                     flavor='sqlite',
                     if_exists=if_exists)

    def indexStat(tablename, column):
        istat = ('create index %(tablename)s_%(column)s '
                 'on %(tablename)s(%(column)s)') % locals()
        return istat

    if index:

        cc = dbhandle.cursor()

        if isinstance(index, str):
            istat = indexStat(tablename, index)
            print(istat)
            db_execute(cc, istat)
        elif isinstance(index, (tuple, list)):
            for column in index:
                istat = indexStat(tablename, column)
                db_execute(cc, istat)

        cc.close()
